#!/usr/bin/env python3
"""
Reads a correct swagger file and produces Elm client code.
"""
import argparse
import pathlib
import json
from typing import TextIO, cast

import swagger_to.elm_client
import swagger_to.intermediate
import swagger_to.swagger


def main() -> None:
    """"
    Main routine
    """
    # pylint: disable=too-many-locals
    parser = argparse.ArgumentParser("Reads a correct swagger file and produces Elm client code.")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--outdir", help="path to the output directory", required=True)
    parser.add_argument("--force", help="overwrite existing files", action="store_true")
    args = parser.parse_args()

    assert isinstance(args.force, bool)
    assert isinstance(args.outdir, str)
    assert isinstance(args.swagger_path, str)

    swagger_path = pathlib.Path(args.swagger_path)
    outdir = pathlib.Path(args.outdir)
    force = bool(args.force)

    if not swagger_path.exists():
        raise FileNotFoundError("Swagger file does not exist: {}".format(swagger_path))

    if not outdir.exists():
        pathlib.Path.mkdir(outdir)

    if not force:
        for fname in ['Client.elm', 'elm-package.sample.json']:
            pth = outdir / fname
            if pth.exists():
                raise FileExistsError("File exists, but --force was not specified: {!r}".format(pth))

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=swagger_path)
    if errs:
        raise ValueError("Failed to parse Swagger file {}:\n{}".format(swagger_path, "\n".join(errs)))

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)

    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    elm_typedefs = swagger_to.elm_client.to_typedefs(intermediate_typedefs=intermediate_typedefs)
    elm_requests = swagger_to.elm_client.to_requests(endpoints=endpoints, typedefs=elm_typedefs)

    src_pth = outdir / 'Client.elm'
    with src_pth.open('wt') as fid:
        fid_textio = cast(TextIO, fid)
        swagger_to.elm_client.write_client_elm(typedefs=elm_typedefs, requests=elm_requests, fid=fid_textio)

    pkg_pth = outdir / 'elm-package.sample.json'
    needs_query_strings = swagger_to.elm_client.needs_query_strings(requests=elm_requests)
    elm_package_json = swagger_to.elm_client.elm_package_json(query_strings=needs_query_strings)
    with pkg_pth.open('wt') as fid:
        json.dump(elm_package_json, fp=fid, indent=2, sort_keys=False)

    print("Generated Elm client code in: {}".format(outdir))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Generates code for an Elm client.
"""
# pylint: disable=too-many-lines
from typing import Optional, MutableMapping, List, Dict, TextIO, Any  # pylint: disable=unused-import
import collections

import swagger_to.intermediate
import swagger_to

INDENT = ' ' * 4


class Typedef:
    """
    Represents an Elm type.
    """

    def __init__(self) -> None:
        self.description = ''
        self.identifier = ''

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.identifier)


class BracketedTypedef(Typedef):
    """
    Represents an Elm type wrapped in brackets.
    """

    def __str__(self) -> str:
        return '({})'.format(self.__class__.__name__)


class Booldef(Typedef):
    """
    Represents Elm booleans.
    """
    pass


class Intdef(Typedef):
    """
    Represents Elm Ints.
    """
    pass


class Floatdef(Typedef):
    """
    Represents Elm floating-point numbers.
    """
    pass


class Stringdef(Typedef):
    """
    Represents Elm strings.
    """
    pass


class Listdef(BracketedTypedef):
    """
    Represents Elm lists.
    """

    def __init__(self) -> None:
        self.items = None  # type: Optional[Typedef]
        super().__init__()


class Dictdef(BracketedTypedef):
    """
    Represents Elm dicts.
    """

    def __init__(self) -> None:
        self.values = None  # type: Optional[Typedef]
        super().__init__()


class Property:
    """
    Represents an Elm record property.
    """

    def __init__(self) -> None:
        self.name = ''
        self.description = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False

    def __str__(self) -> str:
        return '{}: {}'.format(self.name, self.typedef)


class Recorddef(Typedef):
    """
    Represents Elm records.
    """

    def __init__(self) -> None:
        self.properties = collections.OrderedDict()  # type: MutableMapping[str, Property]
        super().__init__()


class Parameter:
    """
    Represents a parameter of a request.
    """

    def __init__(self) -> None:
        self.name = ''
        self.typedef = None  # type: Optional[Typedef]
        self.required = False


class Response:
    """
    Represents a response from a request.
    """

    def __init__(self) -> None:
        self.code = ''
        self.typedef = None  # type: Optional[Typedef]
        self.description = ''


class Request:
    """
    Represents a request function of the client.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self) -> None:
        self.operation_id = ''
        self.path = ''
        self.method = ''
        self.description = ''

        # the union of all parameters
        self.parameters = []  # type: List[Parameter]

        # parameters split into categories
        self.body_parameter = None  # type: Optional[Parameter]
        self.query_parameters = []  # type: List[Parameter]
        self.path_parameters = []  # type: List[Parameter]

        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]


def to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    """
    Translates the intermediate typedef to an Elm typedef.

    :param intermediate_typedef: to be translated
    :return:
    """
    # pylint: disable=too-many-branches
    typedef = None  # type: Optional[Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        if intermediate_typedef.type == 'boolean':
            typedef = Booldef()
        elif intermediate_typedef.type == "integer":
            if intermediate_typedef.format in ["", "int32", "int64"]:
                typedef = Intdef()
            else:
                raise NotImplementedError("Unhandled format of a swagger intermediate type 'integer': {}".format(
                    intermediate_typedef.format))
        elif intermediate_typedef.type == 'number':
            typedef = Floatdef()
        elif intermediate_typedef.type == 'string':
            typedef = Stringdef()
        else:
            raise NotImplementedError("Converting intermediate type to Elm is not supported: {}".format(
                intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        typedef = Listdef()
        typedef.items = to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Dictdef()
        typedef.values = to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Recorddef()

        for intermediate_prop in intermediate_typedef.properties.values():
            prop = Property()
            prop.description = intermediate_prop.description
            prop.name = intermediate_prop.name
            prop.typedef = to_typedef(intermediate_typedef=intermediate_prop.typedef)
            prop.required = intermediate_prop.required

            typedef.properties[prop.name] = prop
    else:
        raise NotImplementedError("Converting intermediate typedef to Elm is not supported: {!r}".format(
            type(intermediate_typedef)))

    typedef.description = intermediate_typedef.description
    typedef.identifier = intermediate_typedef.identifier

    assert typedef is not None

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    """
    Translates the intermediate typedefs to Elm typedefs.

    :param intermediate_typedefs: to be translated
    :return: translated typedefs
    """
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                             typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    If the type has an identifier, it is retrieved from the translated typedefs.

    Otherwise, it is translated on the spot to the corresponding Elm type.

    :param intermediate_typedef: to be translated
    :param typedefs: translated type definitions
    :return: translated type
    """
    if intermediate_typedef.identifier:
        if intermediate_typedef.identifier not in typedefs:
            raise KeyError("Intermediate typedef not found in the translated typedefs: {!r}".format(
                intermediate_typedef.identifier))

        return typedefs[intermediate_typedef.identifier]

    return to_typedef(intermediate_typedef=intermediate_typedef)


def to_parameter(intermediate_parameter: swagger_to.intermediate.Parameter,
                 typedefs: MutableMapping[str, Typedef]) -> Parameter:
    """
    Translates an intermediate parameter to an Elm parameter.

    :param intermediate_parameter: to be translated
    :param typedefs: translated type definitions
    :return: translated parameter
    """
    param = Parameter()
    param.name = intermediate_parameter.name
    param.typedef = anonymous_or_get_typedef(intermediate_typedef=intermediate_parameter.typedef, typedefs=typedefs)
    param.required = intermediate_parameter.required
    return param


def to_response(intermediate_response: swagger_to.intermediate.Response,
                typedefs: MutableMapping[str, Typedef]) -> Response:
    """
    Translates an intermediate response to an Elm response.

    :param intermediate_response: to be translated
    :param typedefs: translated type definitions
    :return: translated response
    """
    resp = Response()
    resp.code = intermediate_response.code
    resp.typedef = None if intermediate_response.typedef is None else \
        anonymous_or_get_typedef(intermediate_typedef=intermediate_response.typedef, typedefs=typedefs)
    resp.description = intermediate_response.description
    return resp


def to_request(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Request:
    """
    Translates an endpoint to a client request function.

    :param endpoint: to be translated
    :param typedefs: translated type definitions
    :return: request function
    """
    if endpoint.produces != ['application/json']:
        raise ValueError("Can not translate an end point to Elm client "
                         "which does not produces strictly application/json: {} {}".format(
                             endpoint.path, endpoint.method))

    req = Request()
    req.description = endpoint.description
    req.method = endpoint.method
    req.operation_id = endpoint.operation_id
    req.path = endpoint.path

    for intermediate_param in endpoint.parameters:
        param = to_parameter(intermediate_parameter=intermediate_param, typedefs=typedefs)

        if intermediate_param.in_what == 'body':
            if req.body_parameter is not None:
                raise KeyError("Duplicate body parameters in an endpoint: {} {}".format(
                    req.body_parameter.name, intermediate_param.name))

            req.body_parameter = param
        elif intermediate_param.in_what == 'query':
            req.query_parameters.append(param)
        elif intermediate_param.in_what == 'path':
            req.path_parameters.append(param)
        else:
            raise NotImplementedError("Unsupported parameter 'in' to Elm translation: {}".format(
                intermediate_param.in_what))

        req.parameters.append(param)

    # parameters are sorted so that first come the required ones; Elm requires the optional ones to come
    # at the end.
    req.parameters.sort(key=lambda param: not param.required)

    for code, intermediate_resp in endpoint.responses.items():
        req.responses[code] = to_response(intermediate_response=intermediate_resp, typedefs=typedefs)

    return req


def to_requests(endpoints: List[swagger_to.intermediate.Endpoint],
                typedefs: MutableMapping[str, Typedef]) -> List[Request]:
    """
    Translates the endpoints to Elm request functions.

    Endpoints which do not produce strictly 'application/json' are ignored.

    :param endpoints: to be translated
    :param typedefs: translated type definitions
    :return: translated request functions
    """
    requests = []  # type: List[Request]
    for endpoint in endpoints:
        if endpoint.produces != ['application/json']:
            continue

        requests.append(to_request(endpoint=endpoint, typedefs=typedefs))

    return requests


def write_description(description: str, indent: str, fid: TextIO) -> None:
    """
    Writes a description as -- comment block.

    :param description: to be written
    :param indent: indention as prefix to each line
    :param fid: target
    :return:
    """
    lines = description.strip().splitlines()
    for i, line in enumerate(lines):
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            fid.write(indent + '-- {}'.format(rstripped))
        else:
            fid.write(indent + '-- ')

        if i < len(lines) - 1:
            fid.write('\n')


def write_top_level_description(description: str, fid: TextIO) -> None:
    """
    Writes a top level description as {|- -} comment block.

    :param description: to be written
    :param fid: target
    :return:
    """
    lines = description.strip().splitlines()
    fid.write('{-| ')
    for i, line in enumerate(lines):
        if i > 0:
            fid.write('    ')
        fid.write('{}\n'.format(line.rstrip()))

    fid.write('-}')


def argument_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to an argument expression for an Elm function signature.
    Most importantly, this means it add brackets where they are due.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type expression
    """

    if isinstance(typedef, BracketedTypedef):
        return '(' + type_expression(typedef=typedef, path=path) + ')'

    return type_expression(typedef=typedef, path=path)


def argument_decoder_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a argument decoder expression for an Elm function signature.
    Most importantly, this means it add brackets where they are due.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type expression
    """

    if isinstance(typedef, BracketedTypedef):
        return '(' + type_decoder(typedef=typedef, path=path) + ')'

    return type_decoder(typedef=typedef, path=path)


def write_type_definition(typedef: Typedef, fid: TextIO) -> None:
    """
    Writes the type definition in the Elm code.

    :param typedef: to be declared
    :param fid: target
    :return:
    """
    if typedef.identifier == '':
        raise ValueError("Expected a typedef with an identifier, but got a typedef with an empty identifier.")

    if typedef.description:
        write_top_level_description(description=typedef.description, fid=fid)
        fid.write('\n')

    if isinstance(typedef, Recorddef):
        fid.write("type alias {} = \n".format(typedef.identifier))
        prefix = '{ '
        for prop in typedef.properties.values():
            if prop.description:
                write_description(description=prop.description, indent=INDENT, fid=fid)
                fid.write("\n")

            type_expr = type_expression(typedef=prop.typedef, path='{}.{}'.format(typedef.identifier, prop.name))
            camel_case_name = swagger_to.camel_case(identifier=prop.name)
            if not prop.required:
                fid.write(INDENT + prefix + '{} : Maybe ({})\n'.format(camel_case_name, type_expr))
            else:
                fid.write(INDENT + prefix + '{} : {}\n'.format(camel_case_name, type_expr))
            prefix = ', '

        fid.write(INDENT + "}")
    else:
        fid.write("type alias {} = \n".format(typedef.identifier))
        fid.write(INDENT + "{}".format(type_expression(typedef=typedef, path=typedef.identifier)))


def write_encoder(typedef: Typedef, fid: TextIO) -> None:
    """
    Writes the Encoder in the Elm code.

    :param typedef: to be encoded
    :param fid: target
    :return:
    """
    if typedef.identifier == '':
        raise ValueError("Expected a typedef with an identifier, but got a typedef with an empty identifier.")

    if isinstance(typedef, Recorddef):
        obj_name = 'a' + typedef.identifier
        fid.write("encode{0} : {0} -> Json.Encode.Value\n".format(typedef.identifier))
        fid.write("encode{} {} =\n".format(typedef.identifier, obj_name))
        fid.write(INDENT + 'Json.Encode.object\n')
        prefix = 2 * INDENT + '[ '
        for prop in typedef.properties.values():
            fid.write(prefix + '( "{}", '.format(swagger_to.snake_case(identifier=prop.name)))

            encoder = type_encoder(typedef=prop.typedef, path='{}.{}'.format(typedef.identifier, prop.name))
            if not prop.required:
                fid.write('Json.Encode.Extra.maybe ({}) '.format(encoder))
            else:
                fid.write('{} '.format(encoder))
            fid.write('<| {}.{} )\n'.format(obj_name, swagger_to.camel_case(identifier=prop.name)))
            prefix = 2 * INDENT + ', '

        fid.write(2 * INDENT + "]")

    elif isinstance(typedef, (Booldef, Intdef, Floatdef, Stringdef, Listdef, Dictdef)):
        bracketed_type_expression = argument_expression(typedef=typedef, path=typedef.identifier)
        var_name = 'a' + typedef.identifier

        fid.write("encode{} : {} -> Json.Encode.Value \n".format(typedef.identifier, bracketed_type_expression))
        fid.write("encode{} {} =\n".format(typedef.identifier, var_name))
        fid.write(INDENT + "{}".format(type_encoder(typedef=typedef, path=typedef.identifier)))
        fid.write(" <| {}".format(var_name))

    else:
        raise AssertionError("Unexpected type {}.".format(typedef.__class__))


def write_decoder(typedef: Typedef, fid: TextIO) -> None:
    """
    Writes the Decoder in the Elm code.

    :param typedef: to be decoded
    :param fid: target
    :return:
    """
    if typedef.identifier == '':
        raise ValueError("Expected a typedef with an identifier, but got a typedef with an empty identifier.")

    if isinstance(typedef, Recorddef):
        record_id = typedef.identifier
        fid.write("decode{0} : Json.Decode.Decoder {0}\n".format(record_id))
        fid.write("decode{} =\n".format(record_id))
        fid.write(INDENT + 'Json.Decode.Pipeline.decode {}\n'.format(record_id))
        prefix = 2 * INDENT + '|> '

        for prop in typedef.properties.values():
            snake_case_name = swagger_to.snake_case(identifier=prop.name)
            decoder = argument_decoder_expression(typedef=prop.typedef, path='{}.{}'.format(record_id, prop.name))
            if prop.required:
                fid.write(prefix + 'Json.Decode.Pipeline.required "{}" {}\n'.format(snake_case_name, decoder))
            else:
                fid.write(prefix + 'Json.Decode.Pipeline.optional "{}" (Json.Decode.nullable {}) Nothing\n'.format(
                    snake_case_name, decoder))

    elif isinstance(typedef, (Booldef, Intdef, Floatdef, Stringdef, Listdef, Dictdef)):
        bracketed_type_expression = argument_expression(typedef=typedef, path=typedef.identifier)
        fid.write("decode{} : Json.Decode.Decoder {}\n".format(typedef.identifier, bracketed_type_expression))
        fid.write("decode{} =\n".format(typedef.identifier))
        fid.write(INDENT + "{}".format(type_decoder(typedef=typedef, path=typedef.identifier)))

    else:
        raise AssertionError("Unexpected type {}.".format(typedef.__class__))


def escape_string(to_escape: str) -> str:
    """

    :param to_escape: string to escape
    :return: the same string, with all Elm special characters escaped.
    """

    to_escape.replace('\\', '\\\\')
    to_escape.replace('\n', '\\n')
    to_escape.replace('\r', '\\r')
    to_escape.replace('\t', '\\t')
    to_escape.replace("\'", "\\'")
    to_escape.replace('\"', '\\"')

    return to_escape


def write_request(request: Request, fid: TextIO) -> None:
    """
    Generates the code of the request function.

    :param request: function definition
    :param fid: target
    :return:
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches
    description = 'Contains a "{}" request to the endpoint: {}, to be sent with Http.send'.format(
        request.method, request.path)
    if request.description:
        description += '\n\n' + request.description
    write_top_level_description(description=description, fid=fid)
    fid.write('\n')

    types = []  # type List[str]
    names = []  # type List[str]
    for param in request.parameters:
        if param.required:
            types.append(type_expression(typedef=param.typedef))
            names.append(swagger_to.camel_case(identifier=param.name))
        else:
            types.append('Maybe {}'.format(type_expression(typedef=param.typedef)))
            names.append('maybe{}'.format(swagger_to.capital_camel_case(param.name)))

    return_type = ''
    return_type_decoder = ''
    if '200' in request.responses:
        resp = request.responses['200']
        if resp.typedef is not None:
            return_type = argument_expression(typedef=resp.typedef)
            return_type_decoder = argument_decoder_expression(typedef=resp.typedef)

    # function signature and arguments
    function_name = '{}Request'.format(swagger_to.camel_case(identifier=request.operation_id))
    types = ["String", "Maybe Time.Time", "Bool"] + types
    names = ["prefix", "maybeTimeout", "withCredentials"] + names
    types_str = ' -> '.join(types)
    types_str += ' -> '
    names_str = ' '.join(names)
    line1 = function_name + ' : ' + types_str + 'Http.Request {}\n'.format(return_type)
    line2 = function_name + ' ' + names_str + ' ='
    indent = 1

    if len(line1) <= 120:
        fid.write(line1)
        fid.write(line2)
    else:
        fid.write(function_name + ' :\n')
        join_str = '\n' + INDENT + '-> '

        fid.write(INDENT + '{}\n'.format(join_str.join(types)))
        fid.write(INDENT + '-> Http.Request {}\n'.format(return_type))

        if len(line2) > 120:
            fid.write(function_name)
            join_str = '\n' + INDENT
            fid.write('\n' + INDENT + '{}'.format(join_str.join(names)))

            fid.write(' =')
            indent = 2
        else:
            fid.write(line2)
    fid.write('\n')

    name_to_parameters = dict([(param.name, param) for param in request.parameters])

    rel_path = request.path[1:] if request.path.startswith('/') else request.path

    fid.write(INDENT * indent + 'let\n')

    # path parameters
    token_pth = swagger_to.tokenize_path(path=rel_path)
    url_name = 'baseUrl' if request.query_parameters else 'url'
    if not token_pth.parameter_to_token_indices:
        fid.write(INDENT * (indent + 1) + '{} = prefix ++ "{}"\n'.format(url_name, rel_path))
    else:
        fid.write(INDENT * (indent + 1) + '{} = prefix'.format(url_name))
        for i, tkn in enumerate(token_pth.tokens):
            fid.write("\n")

            if i in token_pth.token_index_to_parameter:
                param_name = token_pth.token_index_to_parameter[i]
                param = name_to_parameters[param_name]
                camel_case_name = swagger_to.camel_case(identifier=param.name)
                if not isinstance(param.typedef, Stringdef):
                    camel_case_name = '(toString {})'.format(camel_case_name)

                fid.write(INDENT * (indent + 2) + '++ {}'.format(camel_case_name))
            else:
                # escape special characters
                fid.write(INDENT * (indent + 2) + '++ "{}"'.format(escape_string(to_escape=tkn)))

    if request.path_parameters and request.query_parameters:
        fid.write("\n")

    # query parameters
    if request.query_parameters:
        fid.write(INDENT * (indent + 1) + 'queryString = QueryString.empty\n')
        for i, param in enumerate(request.query_parameters):
            camel_case_name = swagger_to.camel_case(identifier=param.name)
            if param.required:
                arg = camel_case_name
                if not isinstance(param.typedef, Stringdef):
                    arg = '(toString {})'.format(camel_case_name)
                fid.write(INDENT * (indent + 2) + '|> QueryString.add "{}" {}\n'.format(param.name, arg))
            else:
                var_name = 'maybe{}'.format(swagger_to.capital_camel_case(identifier=param.name))
                fid.write(INDENT * (indent + 2) + '|> \\queryStr -> Maybe.withDefault queryStr (Maybe.map \n')
                if isinstance(param.typedef, Stringdef):
                    function_param = 'var'
                else:
                    function_param = '(toString var)'
                fid.write(INDENT * (indent + 3) + '(\\var -> QueryString.add "{}" {} queryStr) {})\n'.format(
                    param.name, function_param, var_name))

        fid.write(INDENT * (indent + 1) + 'url = baseUrl ++ (QueryString.render queryString)\n')

    fid.write(indent * INDENT + 'in\n')

    mth = request.method.upper()

    fid.write(INDENT * indent + 'Http.request\n')
    fid.write(INDENT * (indent + 1) + '{ body = ')
    if request.body_parameter is not None:
        if not request.body_parameter.required:
            fid.write('Json.Encode.Extra.maybe ({}'.format(type_encoder(request.body_parameter.typedef)))
        fid.write('({}'.format(type_encoder(request.body_parameter.typedef)))

        fid.write(' {}) |> Http.jsonBody'.format(swagger_to.camel_case(identifier=request.body_parameter.name)))
    else:
        fid.write('Http.emptyBody')
    fid.write('\n')
    fid.write(INDENT * (indent + 1) + ', expect = Http.expectJson {}\n'.format(return_type_decoder))
    fid.write(INDENT * (indent + 1) + ', headers = []\n')
    fid.write(INDENT * (indent + 1) + ', method = "{}"\n'.format(mth))
    fid.write(INDENT * (indent + 1) + ', timeout = maybeTimeout\n')
    fid.write(INDENT * (indent + 1) + ', url = url\n')
    fid.write(INDENT * (indent + 1) + ', withCredentials = withCredentials\n')
    fid.write(INDENT * (indent + 1) + '}\n')


def write_type_definitions(typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Writes all type definitions as Elm code.

    :param typedefs: type definitions to be declared
    :param fid: target
    :return:
    """
    assert typedefs
    fid.write('-- Models\n\n\n')
    for i, typedef in enumerate(typedefs.values()):
        if i > 0:
            fid.write('\n\n\n')

        write_type_definition(typedef=typedef, fid=fid)
    fid.write('\n\n\n')


def write_encoders(typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Writes all JSON encoders as Elm code.

    :param typedefs: type definitions to be encoded
    :param fid: target
    :return:
    """
    assert typedefs
    fid.write('-- Encoders\n\n\n')
    for i, typedef in enumerate(typedefs.values()):
        if i > 0:
            fid.write('\n\n\n')

        write_encoder(typedef=typedef, fid=fid)

    fid.write('\n\n\n')


def write_decoders(typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Writes all JSON decoders as Elm code.

    :param typedefs: type definitions to be decoded
    :param fid: target
    :return:
    """
    assert typedefs
    fid.write('-- Decoders\n\n\n')
    for i, typedef in enumerate(typedefs.values()):
        if i > 0:
            fid.write('\n\n\n')

        write_decoder(typedef=typedef, fid=fid)

    fid.write('\n\n\n')


def write_client(requests: List[Request], fid: TextIO) -> None:
    """
    Generates the client.

    :param requests: translated request functions
    :param fid: target
    :return:
    """
    assert requests
    fid.write("-- Remote Calls\n\n")

    for request in requests:
        fid.write('\n\n')
        write_request(request=request, fid=fid)

    fid.write("\n\n\n")


def write_header(fid: TextIO, typedefs: MutableMapping[str, Typedef], requests: List[Request]) -> None:
    """
    Writes the header.

    :param fid: target
    :param typedefs: translated type definitions
    :param requests: translated request functions
    :return:
    """
    fid.write("-- Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n\n")

    to_expose = []  # Type: List[str]

    for typedef in typedefs.values():
        to_expose.append(typedef.identifier)
        to_expose.append('decode{}'.format(typedef.identifier))
        to_expose.append('encode{}'.format(typedef.identifier))

    for request in requests:
        to_expose.append('{}Request'.format(swagger_to.camel_case(identifier=request.operation_id)))

    to_expose.sort()
    joinstr = '\n' + INDENT * 2 + ', '
    fid.write("module Client \n")
    fid.write(INDENT + "exposing\n")
    fid.write(INDENT * 2 + "( {}\n".format(joinstr.join(to_expose)))
    fid.write(INDENT * 2 + ")\n\n")
    fid.write("import Dict exposing (Dict)\n"
              "import Http\n"
              "import Json.Decode\n"
              "import Json.Decode.Pipeline\n"
              "import Json.Encode\n"
              "import Json.Encode.Extra\n")

    # import QueryString only if needed
    for request in requests:
        if request.query_parameters:
            fid.write("import QueryString\n")
            break

    fid.write("import Time\n\n\n")


def write_footer(fid: TextIO) -> None:
    """
    Writes the footer (same for all the Elm clients).

    :param fid: target
    :return:
    """

    fid.write("-- Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")


def type_expression(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a type expression.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type expression
    """
    # pylint: disable=too-many-return-statements

    if isinstance(typedef, Booldef):
        return 'Bool'
    elif isinstance(typedef, Floatdef):
        return 'Float'
    elif isinstance(typedef, Intdef):
        return 'Int'
    elif isinstance(typedef, Stringdef):
        return 'String'
    elif isinstance(typedef, Listdef):
        return 'List ' + type_expression(typedef=typedef.items, path=str(path) + '.items')
    elif isinstance(typedef, Dictdef):
        return 'Dict String ' + type_expression(typedef=typedef.values, path=str(path) + '.values')
    elif isinstance(typedef, Recorddef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to an Elm type expression is not supported: {}".format(path))

        return typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to a type expression is not supported: {!r}: {}".format(
            type(typedef), path))


def type_encoder(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a type encoder.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type encoder
    """
    # pylint: disable=too-many-return-statements

    if isinstance(typedef, Booldef):
        return 'Json.Encode.bool'
    elif isinstance(typedef, Floatdef):
        return 'Json.Encode.float'
    elif isinstance(typedef, Intdef):
        return 'Json.Encode.int'
    elif isinstance(typedef, Stringdef):
        return 'Json.Encode.string'
    elif isinstance(typedef, Listdef):
        return 'Json.Encode.list <| List.map ' + type_encoder(typedef=typedef.items, path=str(path) + '.items')
    elif isinstance(typedef, Dictdef):
        return 'Json.Encode.Extra.dict identity ' + type_encoder(typedef=typedef.values, path=str(path) + '.values')
    elif isinstance(typedef, Recorddef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to an Elm type expression is not supported: {}".format(path))

        return 'encode' + typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to an encoder is not supported: {!r}: {}".format(
            type(typedef), path))


def type_decoder(typedef: Typedef, path: Optional[str] = None) -> str:
    """
    Translates the typedef to a type decoder.

    :param typedef: to be translated
    :param path: path in the intermediate representation
    :return: type decoder
    """
    # pylint: disable=too-many-return-statements

    if isinstance(typedef, Booldef):
        return 'Json.Decode.bool'
    elif isinstance(typedef, Floatdef):
        return 'Json.Decode.float'
    elif isinstance(typedef, Intdef):
        return 'Json.Decode.int'
    elif isinstance(typedef, Stringdef):
        return 'Json.Decode.string'
    elif isinstance(typedef, Listdef):
        return 'Json.Decode.list <| ' + type_decoder(typedef=typedef.items, path=str(path) + '.items')
    elif isinstance(typedef, Dictdef):
        return 'Json.Decode.dict <| ' + type_decoder(typedef=typedef.values, path=str(path) + '.values')
    elif isinstance(typedef, Recorddef):
        if typedef.identifier == '':
            raise NotImplementedError(
                "Translating an anonymous class to an Elm type expression is not supported: {}".format(path))

        return 'decode' + typedef.identifier
    else:
        raise NotImplementedError("Translating the typedef to a decoder is not supported: {!r}: {}".format(
            type(typedef), path))


def needs_query_strings(requests: List[Request]) -> bool:
    """
    Determines whether the QueryString package is needed by the client.

    :param requests: translated request functions.
    :return: true if any of the Requests contains nonempty query parameters, false otherwise
    """
    for request in requests:
        if request.query_parameters:
            return True

    return False


def elm_package_json(query_strings: bool) -> MutableMapping[str, Any]:
    """
    Returns the elm-package json file.

    :param query_strings: if true, the package Bogdanp/querystring is inserted in the package.
    :return: The JSON Elm package for the project as a Dictionary.
    """
    elm_pkg = collections.OrderedDict()  # type: Dict[str, Any]
    elm_pkg['version'] = '1.1.0'
    elm_pkg['summary'] = ''
    elm_pkg['repository'] = 'https://github.com/invalid/repo.git'
    elm_pkg['license'] = ''
    elm_pkg['source-directories'] = ['src']
    elm_pkg['exposed-modules'] = []

    packages = collections.OrderedDict()  # type: Dict[str, str]
    packages['elm-lang/core'] = '2.0.0 <= v <= 2.0.0'
    packages['elm-lang/http'] = '1.0.0 <= v <= 1.0.0'
    packages['elm-community/json-extra'] = '2.7.0 <= v <= 2.7.0'
    if query_strings:
        packages['Bogdanp/elm-querystring'] = '1.0.0 <= v <= 1.0.0'
    packages['NoRedInk/elm-decode-pipeline'] = '3.0.0 <= v <= 3.0.0'

    elm_pkg['dependencies'] = packages
    elm_pkg['elm-version'] = '0.18.0 <= v < 0.19.0'

    return elm_pkg


def write_client_elm(typedefs: MutableMapping[str, Typedef], requests: List[Request], fid: TextIO) -> None:
    """
    Generates the file with the client code.

    :param typedefs: translated type definitions
    :param requests: translated request functions
    :param fid: target
    :return:
    """
    write_header(fid=fid, typedefs=typedefs, requests=requests)

    if typedefs:
        write_type_definitions(typedefs=typedefs, fid=fid)

        write_encoders(typedefs=typedefs, fid=fid)

        write_decoders(typedefs=typedefs, fid=fid)

    if requests:
        write_client(requests=requests, fid=fid)

    write_footer(fid=fid)



"""
Swagger representation
"""
import pathlib

import collections
import sys
from typing import List, Optional, MutableMapping, Any, Tuple, Union  # pylint: disable=unused-import

try:
    import yaml
    from yaml.composer import Composer
    from yaml.constructor import Constructor
except ImportError as err:
    print("Package pyyaml was not installed. pip3 install pyyaml?")
    sys.exit(1)
# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches


class RawDict:
    """Represents a raw dictionary from a schema file."""

    def __init__(self, adict=collections.OrderedDict(), source='', lineno=0):
        self.adict = adict
        self.source = source
        self.lineno = lineno


class Typedef:
    def __init__(self):
        self.ref = ''
        self.description = ''
        self.type = ''
        self.format = ''
        self.pattern = ''
        self.properties = collections.OrderedDict()  # type: MutableMapping[str, Typedef]
        self.required = []  # type: List[str]
        self.items = None  # type: Optional[Typedef]
        self.additional_properties = None  # type: Optional[Typedef]
        self.__lineno__ = 0

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Definition:
    def __init__(self):
        self.identifier = ''
        self.typedef = None  # type: Optional[Typedef]
        self.swagger = None  # type: Optional[Swagger]

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Parameter:
    def __init__(self):
        self.method = None  # type: Optional[Method]
        self.name = ''
        self.in_what = ''
        self.description = ''
        self.required = False
        self.type = ''
        self.format = ''
        self.pattern = ''
        self.schema = None  # type: Optional[Typedef]
        self.ref = ''
        self.__lineno__ = 0

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Response:
    def __init__(self):
        self.code = ''
        self.description = ''
        self.schema = None  # type: Optional[Typedef]
        self.type = ''
        self.format = ''
        self.pattern = ''
        self.__lineno__ = 0

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Method:
    def __init__(self):
        self.identifier = ''
        self.operation_id = ''
        self.tags = []  # type: List[str]
        self.description = ''
        self.parameters = []  # type: List[Parameter]
        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]
        self.path = None  # type: Optional[Path]
        self.produces = []  # type: List[str]
        self.consumes = []  # type: List[str]
        self.x_swagger_to_skip = False
        self.__lineno__ = 0

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Path:
    def __init__(self):
        self.identifier = ''
        self.methods = []  # type: List[Method]
        self.swagger = None  # type: Optional[Swagger]
        self.__lineno__ = 0

        # original specification dictionary, if available; not deep-copied, do not modify
        self.raw_dict = None  # type: RawDict


class Swagger:
    def __init__(self):
        self.name = ""
        self.base_path = ""
        self.description = ""
        self.paths = collections.OrderedDict()  # type: MutableMapping[str, Path]
        self.definitions = collections.OrderedDict()  # type: MutableMapping[str, Definition]
        self.parameters = collections.OrderedDict()  # type: MutableMapping[str, Parameter]

        self.raw_dict = None  # type: RawDict


def parse_typedef(raw_dict: RawDict) -> Tuple[Typedef, List[str]]:
    adict = raw_dict.adict

    typedef = Typedef()
    typedef.ref = adict.get('$ref', '')
    typedef.description = adict.get('description', '').strip()
    typedef.type = adict.get('type', '')
    typedef.format = adict.get('format', '')
    typedef.pattern = adict.get('pattern', '')
    typedef.__lineno__ = raw_dict.lineno

    errors = []  # type: List[str]

    for prop_name, prop_dict in adict.get('properties', RawDict()).adict.items():
        prop_typedef, prop_errors = parse_typedef(raw_dict=prop_dict)

        errors.extend(['in property {!r}: {}'.format(prop_name, error) for error in prop_errors])
        typedef.properties[prop_name] = prop_typedef

    typedef.required = adict.get('required', [])

    # check that all the required are well-defined
    for prop_name in typedef.required:
        if prop_name not in typedef.properties:
            errors.append("required property not defined: {!r}".format(prop_name))

    if 'additionalProperties' in adict:
        add_prop_dict = adict['additionalProperties']
        add_prop_typedef, add_prop_errors = parse_typedef(raw_dict=add_prop_dict)

        errors.extend(['in additionalProperties: {}'.format(error) for error in add_prop_errors])
        typedef.additional_properties = add_prop_typedef

    if 'items' in adict:
        items_dict = adict['items']
        items_typedef, items_errors = parse_typedef(raw_dict=items_dict)

        errors.extend(['in items: {}'.format(error) for error in items_errors])
        typedef.items = items_typedef

    if typedef.type == 'number':
        if typedef.format not in ['float', 'double']:
            errors.append("Unexpected format for type 'number': {!r}".format(typedef.format))

    elif typedef.type == 'integer':
        if typedef.format not in ['int32', 'int64']:
            errors.append("Unexpected format for type 'integer': {!r}".format(typedef.format))

    typedef.raw_dict = raw_dict

    return typedef, errors


def parse_parameter(raw_dict: RawDict) -> Tuple[Parameter, List[str]]:
    """
    Parses a parameter from the dictionary.

    :param raw_dict: to be parsed
    :return: parameter, list of errors
    """
    adict = raw_dict.adict

    param = Parameter()
    param.name = adict.get('name', '')
    param.in_what = adict.get('in', '')
    param.description = adict.get('description', '').strip()
    param.required = adict.get('required', False)
    param.type = adict.get('type', '')
    param.format = adict.get('format', '')
    param.pattern = adict.get('pattern', '')
    param.ref = adict.get('$ref', '')
    param.__lineno__ = raw_dict.lineno

    errors = []  # type: List[str]

    if 'schema' in adict:
        schema_dict = adict['schema']

        typedef, schema_errors = parse_typedef(raw_dict=schema_dict)
        param.schema = typedef
        errors.extend(['in schema: {}'.format(error) for error in schema_errors])

    param.raw_dict = raw_dict

    if param.in_what == 'body' and param.schema is None:
        errors.append('parameter in body, but no schema')

    if 'default' in adict:
        errors.append('default values for parameters are not supported')

    return param, errors


def parse_response(raw_dict: RawDict) -> Tuple[Response, List[str]]:
    """
    Parses a response from the dictionary.

    :param raw_dict: to be parsed
    :return: response, list of errors
    """
    adict = raw_dict.adict

    resp = Response()
    errors = []  # type: List[str]

    resp.description = adict.get('description', '').strip()
    resp.type = adict.get('type', '')
    resp.format = adict.get('format', '')
    resp.pattern = adict.get('pattern', '')
    resp.__lineno__ = raw_dict.lineno

    if 'schema' in adict:
        schema_dict = adict['schema']

        typedef, schema_errors = parse_typedef(raw_dict=schema_dict)
        resp.schema = typedef
        errors.extend(['in schema: {}'.format(error) for error in schema_errors])

    resp.raw_dict = raw_dict

    return resp, errors


def parse_method(raw_dict: RawDict) -> Tuple[Method, List[str]]:
    """
    Parses a method from the dictionary.

    :param raw_dict: to be parsed
    :return: method, list of errors
    """
    mth = Method()
    errors = []  # type: List[str]
    adict = raw_dict.adict

    mth.operation_id = adict.get('operationId', '')
    if mth.operation_id == '':
        errors.append('missing operationId')

    mth.tags = adict.get('tags', [])
    mth.description = adict.get('description', '').strip()
    mth.x_swagger_to_skip = adict.get('x-swagger-to-skip', False)

    mth.produces = adict.get('produces', [])
    mth.consumes = adict.get('consumes', [])
    mth.__lineno__ = raw_dict.lineno

    for i, param_dict in enumerate(adict.get('parameters', [])):
        param, param_errors = parse_parameter(raw_dict=param_dict)
        errors.extend(['in parameter {} (name: {!r}): {}'.format(i, param.name, error) for error in param_errors])

        param.method = mth

        mth.parameters.append(param)

    for resp_code, resp_dict in adict.get('responses', RawDict()).adict.items():
        resp, resp_errors = parse_response(raw_dict=resp_dict)
        errors.extend(['in response {!r}: {}'.format(resp_code, error) for error in resp_errors])

        resp.code = resp_code
        mth.responses[str(resp_code)] = resp

    mth.raw_dict = raw_dict

    return mth, errors


def parse_path(raw_dict: RawDict) -> Tuple[Path, List[str]]:
    """
    Parses a path from the dictionary.

    :param path_id: identifier
    :param raw_dict: to be parsed
    :return: path, list of errors
    """
    pth = Path()
    errors = []  # type: List[str]

    for method_id, method_dict in raw_dict.adict.items():
        method, method_errors = parse_method(raw_dict=method_dict)
        method.identifier = method_id
        method.path = pth
        errors.extend(['in method {!r}: {}'.format(method_id, error) for error in method_errors])

        if not method_errors:
            pth.methods.append(method)

    pth.raw_dict = raw_dict

    return pth, errors


def parse_yaml(stream: Any) -> Tuple[Swagger, List[str]]:
    """
    Parses the Swagger specification from the given text.

    :param stream: YAML representation of the Swagger spec satisfying file interface
    :return: Swagger specification, list of errors
    """
    # adapted from https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    # and https://stackoverflow.com/questions/13319067/parsing-yaml-return-with-line-number

    object_pairs_hook = collections.OrderedDict

    class OrderedLoader(yaml.SafeLoader):
        def compose_node(self, parent, index):
            # the line number where the previous token has ended (plus empty lines)
            node = Composer.compose_node(self, parent, index)
            node.__lineno__ = self.line + 1
            return node

    def construct_mapping(loader, node, deep=False):
        loader.flatten_mapping(node)
        mapping = Constructor.construct_pairs(loader, node, deep=deep)

        ordered_hook = object_pairs_hook(mapping)

        # assert not hasattr(ordered_hook, "__lineno__"), \
        #     "Expected ordered mapping to have no __lineno__ attribute set before"
        # setattr(ordered_hook, "__lineno__", node.__lineno__)

        return RawDict(adict=ordered_hook, source=stream.name, lineno=node.__lineno__)

    OrderedLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)

    raw_dict = yaml.load(stream, OrderedLoader)
    swagger = Swagger()

    errors = []  # type: List[str]

    adict = raw_dict.adict
    if 'tags' in adict:
        if len(adict['tags']) > 0:
            for tag in adict['tags']:
                for key, value in tag.adict.items():
                    if key == 'name':
                        swagger.name = value

    if swagger.name == '':
        errors.append('missing tag "name" in the swagger specification')

    swagger.base_path = adict.get('basePath', '')

    for path_id, path_dict in adict.get('paths', RawDict()).adict.items():
        path, path_errors = parse_path(raw_dict=path_dict)
        path.identifier = path_id
        path.swagger = swagger

        errors.extend(['in path {!r}: {}'.format(path_id, error) for error in path_errors])

        if not path_errors:
            swagger.paths[path_id] = path

    for def_id, def_dict in adict.get('definitions', RawDict()).adict.items():
        typedef, def_errors = parse_typedef(raw_dict=def_dict)

        errors.extend(['in definition {!r}: {}'.format(def_id, error) for error in def_errors])

        adef = Definition()
        adef.swagger = swagger
        adef.identifier = def_id
        adef.typedef = typedef

        if not def_errors:
            swagger.definitions[def_id] = adef

    for param_id, param_dict in adict.get('parameters', RawDict()).adict.items():
        param, param_errors = parse_parameter(raw_dict=param_dict)

        errors.extend(['in parameter {!r}: {}'.format(param_id, error) for error in param_errors])

        if not param_errors:
            swagger.parameters[param_id] = param

    swagger.raw_dict = raw_dict

    return swagger, errors


def parse_yaml_file(path: Union[str, pathlib.Path]) -> Tuple[Swagger, List[str]]:
    """
    Parses the Swagger specification from the given file.

    :param path: to the .yaml file
    :return: Swagger specification, list of errors
    """
    with open(str(path), 'rt') as fid:
        return parse_yaml(stream=fid)



"""
Parses the Swagger specification to an intermediate representation. This intermediate representation should be easier
to translate to code than a direct Swagger spec translation.
For example, references are resolved and do not figure as strings.
"""

# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches
# pylint: disable=too-many-statements

import collections
import json
from typing import List, MutableMapping, Union, Any, Optional  # pylint: disable=unused-import

import swagger_to.swagger

# see https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md for a list of types
PRIMITIVE_SWAGGER_TYPES = ['string', 'number', 'integer', 'boolean', 'file']


class Typedef:
    def __init__(self):
        self.identifier = ''
        self.description = ''
        self.json_schema = JsonSchema()
        self.line = 0


class Propertydef:
    def __init__(self):
        self.name = ''
        self.typedef = Typedef()
        self.description = ''
        self.required = False
        self.line = 0


class Objectdef(Typedef):
    def __init__(self):
        super().__init__()

        self.properties = collections.OrderedDict()  # type: MutableMapping[str, Propertydef]
        self.required = []  # type: List[str]


class Arraydef(Typedef):
    def __init__(self):
        super().__init__()

        self.items = Typedef()


class Mapdef(Typedef):
    def __init__(self):
        super().__init__()

        self.values = Typedef()


class Primitivedef(Typedef):
    def __init__(self):
        super().__init__()

        self.type = ''
        self.format = ''
        self.pattern = ''


class JsonSchema:
    def __init__(self):
        self.identifier = ''
        self.text = ''


class Parameter:
    def __init__(self):
        self.name = ''
        self.in_what = ''
        self.typedef = Typedef()
        self.required = False
        self.json_schema = JsonSchema()
        self.description = None  # type: Optional[str]
        self.line = 0


class Response:
    def __init__(self) -> None:
        self.code = ''
        self.description = ''
        self.typedef = None  # type: Optional[Typedef]
        self.line = 0


class Endpoint:
    def __init__(self):
        self.path = ''
        self.method = ''
        self.operation_id = ''
        self.parameters = []  # type: List[Parameter]
        self.description = ''
        self.produces = []  # type: List[str]
        self.consumes = []  # type: List[str]
        self.responses = collections.OrderedDict()  # type: MutableMapping[str, Response]
        self.line = 0


def preallocate_named_typedefs(definition: swagger_to.swagger.Definition,
                               typedefs: MutableMapping[str, Typedef]) -> None:
    """
    Adds an entry in `typedefs` with the correct instance of the type definition.

    :param definition: original representation of the definition
    :param typedefs: named intermediate representations of type definitions
    :return:
    """
    typedef = None  # type: Union[None, Primitivedef, Arraydef, Mapdef, Objectdef]
    if definition.typedef.type in PRIMITIVE_SWAGGER_TYPES:
        typedef = Primitivedef()
        typedef.type = definition.typedef.type
        typedef.format = definition.typedef.format
        typedef.pattern = definition.typedef.pattern

    elif definition.typedef.type == 'array':
        typedef = Arraydef()

    elif definition.typedef.type == 'object':
        if definition.typedef.additional_properties is not None:
            typedef = Mapdef()

        else:
            typedef = Objectdef()
            typedef.required = definition.typedef.required
    else:
        raise ValueError("Unexpected type of a typedef in the definition {!r}: {!r}".format(
            definition.identifier, definition.typedef.type))

    typedef.identifier = definition.identifier
    typedef.line = definition.typedef.__lineno__
    typedef.description = definition.typedef.description
    typedefs[typedef.identifier] = typedef


def resolve_substructures(definition: swagger_to.swagger.Definition, typedefs: MutableMapping[str, Typedef]) -> None:
    """
    Resolves substructures (such as property typedefs, item typedefs etc.) of a pre-allocated definition typedef.

    :param definition: original representation of the definition
    :param typedefs: named intermediate representations of type definitions
    :return:
    """
    if definition.identifier not in typedefs:
        raise ValueError("The definition has not been previously allocated: {!r}".format(definition.identifier))

    typedef = typedefs[definition.identifier]

    if isinstance(typedef, Primitivedef):
        pass

    elif isinstance(typedef, Arraydef):
        typedef.items = anonymous_or_get_typedef(original_typedef=definition.typedef.items, typedefs=typedefs)

    elif isinstance(typedef, Mapdef):
        typedef.values = anonymous_or_get_typedef(
            original_typedef=definition.typedef.additional_properties, typedefs=typedefs)

    elif isinstance(typedef, Objectdef):
        for prop_name, prop_typedef in definition.typedef.properties.items():
            propdef = Propertydef()
            propdef.name = prop_name
            propdef.typedef = anonymous_or_get_typedef(original_typedef=prop_typedef, typedefs=typedefs)
            propdef.description = prop_typedef.description
            propdef.required = propdef.name in typedef.required

            typedef.properties[prop_name] = propdef
    else:
        raise ValueError("Unexpected type of the definition {!r}: {}".format(typedef.identifier, type(typedef)))


def anonymous_or_get_typedef(original_typedef: swagger_to.swagger.Typedef,
                             typedefs: MutableMapping[str, Typedef]) -> Typedef:
    """
    Creates an anonymous (unnamed) intermediate type definition or resolves to an object pointed by the 'ref' property.

    :param original_typedef: original (Swagger) type definition
    :param typedefs: named intermediate type definitions
    :return: anonymous (created) or named (resolved) intermediate typedefinition
    """
    if original_typedef.ref != '':
        definition_name = swagger_to.parse_definition_ref(ref=original_typedef.ref)

        if definition_name not in typedefs:
            raise ValueError('The definition referenced from {!r} has not been previously defined'.format(
                original_typedef.ref))

        typedef = typedefs[definition_name]
        return typedef

    if original_typedef.type in PRIMITIVE_SWAGGER_TYPES:
        typedef = Primitivedef()
        typedef.type = original_typedef.type
        typedef.format = original_typedef.format
        typedef.pattern = original_typedef.pattern
        typedef.line = original_typedef.__lineno__

    elif original_typedef.type == 'array':
        typedef = Arraydef()

        if original_typedef.items is None:
            raise ValueError("Unexpected None items: {!r}".format(original_typedef.raw_dict.adict))

        typedef.items = anonymous_or_get_typedef(original_typedef=original_typedef.items, typedefs=typedefs)

    elif original_typedef.type == 'object':
        if original_typedef.additional_properties is not None:
            typedef = Mapdef()

            typedef.values = anonymous_or_get_typedef(
                original_typedef=original_typedef.additional_properties, typedefs=typedefs)
        else:
            typedef = Objectdef()
            typedef.required = original_typedef.required

            for prop_name, prop_typedef in typedef.properties.items():
                propdef = Propertydef()
                propdef.typedef = anonymous_or_get_typedef(original_typedef=prop_typedef.typedef, typedefs=typedefs)
                propdef.description = prop_typedef.description
                propdef.name = prop_name
                propdef.required = propdef.name in typedef.required
                propdef.line = propdef.typedef.line

                typedef.properties[prop_name] = propdef

    else:
        raise ValueError("Unexpected definition: {!r}".format(original_typedef.raw_dict.adict))

    assert typedef is not None

    return typedef


def to_typedefs(swagger: swagger_to.swagger.Swagger) -> MutableMapping[str, Typedef]:
    """
    Translates all original (Swagger) definitions into intermediate type definitions.

    :param swagger: Swagger specification
    :return: intermediate type definitions
    """
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    # pre-allocate type definitions in the first pass so that we can resolve
    # all the references to intermediate typedefs.
    for defi in swagger.definitions.values():
        preallocate_named_typedefs(definition=defi, typedefs=typedefs)

    # resolve references given as URIs to object pointers to intermediate type definitions in a second pass.
    for defi in swagger.definitions.values():
        resolve_substructures(definition=defi, typedefs=typedefs)

    for typedef in typedefs.values():
        json_schema_identifier = typedef.identifier

        typedef.json_schema = to_json_schema(
            identifier=json_schema_identifier,
            original_typedef=swagger.definitions[typedef.identifier].typedef,
            definitions=swagger.definitions)

    return typedefs


def collect_referenced_definitions(typedef: swagger_to.swagger.Typedef,
                                   definitions: MutableMapping[str, swagger_to.swagger.Definition]) -> List[str]:
    if typedef.ref != '':
        definition_name = swagger_to.parse_definition_ref(typedef.ref)
        definition = definitions[definition_name]

        return [definition_name] + collect_referenced_definitions(typedef=definition.typedef, definitions=definitions)

    referenced_definitions = []  # type: List[str]

    for prop in typedef.properties.values():
        referenced_definitions.extend(collect_referenced_definitions(typedef=prop, definitions=definitions))

    if typedef.items is not None:
        referenced_definitions.extend(collect_referenced_definitions(typedef=typedef.items, definitions=definitions))

    if typedef.additional_properties is not None:
        referenced_definitions.extend(
            collect_referenced_definitions(typedef=typedef.additional_properties, definitions=definitions))

    return referenced_definitions


def recursively_strip_descriptions(schema_dict: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
    """
    Strips the value if the key is "description" in-place.

    :param schema_dict: JSON schema as a python dictionary.
    :return:
    """
    new_schema_dict = collections.OrderedDict()  # type: MutableMapping[str, Any]

    for key, value in schema_dict.items():
        if key.lower() == 'description':
            if not isinstance(value, str):
                raise ValueError("Expected the value in a schema to be a string, but got: {}".format(type(value)))

            new_schema_dict[key] = value.strip()
        elif isinstance(value, list):
            lst = []  # type: List[Any]
            for item in value:
                if isinstance(item, (dict, collections.OrderedDict)):
                    lst.append(recursively_strip_descriptions(schema_dict=item))
                else:
                    lst.append(item)

            new_schema_dict[key] = lst

        elif isinstance(value, (dict, collections.OrderedDict)):
            new_schema_dict[key] = recursively_strip_descriptions(schema_dict=value)
        elif isinstance(value, swagger_to.swagger.RawDict):
            new_schema_dict[key] = recursively_strip_descriptions(schema_dict=value.adict)
        else:
            new_schema_dict[key] = value

    return new_schema_dict


def to_json_schema(identifier: str, original_typedef: swagger_to.swagger.Typedef,
                   definitions: MutableMapping[str, swagger_to.swagger.Definition]) -> JsonSchema:
    json_schema = JsonSchema()
    json_schema.identifier = identifier

    schema = collections.OrderedDict()  # type: MutableMapping[str, Any]

    schema['title'] = json_schema.identifier
    schema['$schema'] = "http://json-schema.org/draft-04/schema#"

    referenced_definitions = collect_referenced_definitions(
        typedef=original_typedef, definitions=definitions)  # top-down
    referenced_definitions.reverse()  # bottom-up

    schema_definitions = collections.OrderedDict()  # type: MutableMapping[str, Any]

    for definition_name in referenced_definitions:
        if definition_name in schema_definitions:
            continue

        schema_definitions[definition_name] = definitions[definition_name].typedef.raw_dict.adict

    if len(schema_definitions) > 0:
        schema['definitions'] = schema_definitions

    for key, value in original_typedef.raw_dict.adict.items():
        schema[key] = value

    schema = recursively_strip_descriptions(schema_dict=schema)

    json_schema.text = json.dumps(schema, indent=2)

    return json_schema


def to_parameter(original_param: swagger_to.swagger.Parameter, typedefs: MutableMapping[str, Typedef]) -> Parameter:
    """
    Translates an original parameter to an intermediate parameter representation.

    :param original_param: parameter in the original (Swagger) representation
    :param typedefs: intermediate type definitions
    :return: intermediate representation of a parameter
    """
    if original_param.type != '':
        if original_param.type not in PRIMITIVE_SWAGGER_TYPES:
            raise ValueError("Expected type of a parameter to be one of {!r}, but got: {!r}".format(
                PRIMITIVE_SWAGGER_TYPES, original_param.type))

        original_typedef = swagger_to.swagger.Typedef()
        original_typedef.type = original_param.type
        original_typedef.format = original_param.format
        original_typedef.pattern = original_param.pattern
    elif original_param.schema is not None:
        original_typedef = original_param.schema
    else:
        raise ValueError(
            "Could not resolve the type of the parameter, neither 'type' nor 'schema' defined: {!r}".format(
                original_param.raw_dict.adict))

    typedef = anonymous_or_get_typedef(original_typedef=original_typedef, typedefs=typedefs)

    param = Parameter()
    param.typedef = typedef
    param.in_what = original_param.in_what
    param.name = original_param.name
    param.required = original_param.required
    param.description = original_param.description
    param.line = original_param.__lineno__

    return param


def to_parameters(swagger: swagger_to.swagger.Swagger,
                  typedefs: MutableMapping[str, Typedef]) -> MutableMapping[str, Parameter]:
    params = collections.OrderedDict()  # type: MutableMapping[str, Parameter]

    for original_param in swagger.parameters.values():
        if original_param.ref != '':
            raise ValueError("Expected no 'ref' property in a parameter definition {!r}, but got: {!r}".format(
                original_param.name, original_param.raw_dict.adict))

        param = to_parameter(original_param=original_param, typedefs=typedefs)
        params[param.name] = param

    return params


def anonymous_or_get_parameter(original_param: swagger_to.swagger.Parameter, typedefs: MutableMapping[str, Typedef],
                               params: MutableMapping[str, Parameter]) -> Parameter:
    if original_param.ref != '':
        param_ref_name = swagger_to.parse_parameter_ref(ref=original_param.ref)
        if param_ref_name not in params:
            raise ValueError("The parameter referenced by the parameter {!r} has not been defined: {!r}".format(
                original_param.raw_dict.adict, original_param.ref))

        return params[param_ref_name]

    return to_parameter(original_param=original_param, typedefs=typedefs)


def to_response(swagger_response: swagger_to.swagger.Response, typedefs: MutableMapping[str, Typedef]) -> Response:
    resp = Response()
    resp.description = swagger_response.description
    resp.code = swagger_response.code
    resp.line = swagger_response.__lineno__

    swagger_typedef = None  # type: Optional[swagger_to.swagger.Typedef]

    if swagger_response.type != '':
        if swagger_response.type not in PRIMITIVE_SWAGGER_TYPES:
            raise ValueError("Expected type of a parameter to be one of {!r}, but got: {!r}".format(
                PRIMITIVE_SWAGGER_TYPES, swagger_response.type))

        swagger_typedef = swagger_to.swagger.Typedef()
        swagger_typedef.type = swagger_response.type
        swagger_typedef.format = swagger_response.format
        swagger_typedef.pattern = swagger_response.pattern

    elif swagger_response.schema is not None:
        swagger_typedef = swagger_response.schema
    else:
        # no typedef will be set for this response
        pass

    if swagger_typedef is not None:
        resp.typedef = anonymous_or_get_typedef(original_typedef=swagger_typedef, typedefs=typedefs)

    return resp


def to_endpoint(method: swagger_to.swagger.Method, typedefs: MutableMapping[str, Typedef],
                params: MutableMapping[str, Parameter]) -> Endpoint:
    base_path = method.path.swagger.base_path
    if base_path != '':
        pth = '{}/{}'.format(base_path.rstrip('/'), method.path.identifier.lstrip('/'))
    else:
        pth = method.path.identifier

    endpt = Endpoint()
    endpt.path = pth
    endpt.method = method.identifier
    endpt.operation_id = method.operation_id
    endpt.description = method.description
    endpt.consumes = method.consumes
    endpt.produces = method.produces
    endpt.line = method.__lineno__

    for original_param in method.parameters:
        param = anonymous_or_get_parameter(original_param=original_param, typedefs=typedefs, params=params)

        if param.in_what == 'body':
            if param.typedef.identifier == '':
                json_schema_identifier = original_param.method.operation_id + "_" + param.name
            else:
                json_schema_identifier = param.typedef.identifier

            param.json_schema = to_json_schema(
                identifier=json_schema_identifier,
                original_typedef=original_param.schema,
                definitions=method.path.swagger.definitions)

        endpt.parameters.append(param)

    for resp_code, swagger_resp in method.responses.items():
        endpt.responses[resp_code] = to_response(swagger_response=swagger_resp, typedefs=typedefs)

    return endpt


def to_endpoints(swagger: swagger_to.swagger.Swagger, typedefs: MutableMapping[str, Typedef],
                 params: MutableMapping[str, Parameter]) -> List[Endpoint]:
    endpoints = []  # type: List[Endpoint]
    for path in swagger.paths.values():
        for method in path.methods:
            if not method.x_swagger_to_skip:
                endpoint = to_endpoint(method=method, typedefs=typedefs, params=params)
                endpoints.append(endpoint)

    return endpoints



#!/usr/bin/env python3
"""
Reads a correct swagger file and checks whether it conforms to a style guide.
"""
import argparse
import pathlib
from typing import List

import sys

import swagger_to.intermediate
import swagger_to.style
import swagger_to.swagger


def main() -> int:
    """"
    Main routine
    """
    parser = argparse.ArgumentParser("Reads a correct swagger file and checks that it conforms to the style guide.")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--verbose", help="if set, prints as much information as possible.", action="store_true")
    parser.add_argument(
        "--with_line_number",
        help="if set, prints the errors with the corresponding file name and line number.",
        action="store_true")
    args = parser.parse_args()

    assert isinstance(args.swagger_path, str)
    assert isinstance(args.verbose, bool)
    assert isinstance(args.with_line_number, bool)

    swagger_path = pathlib.Path(args.swagger_path)

    if not swagger_path.exists():
        print("File not found error: Swagger file does not exist: {}".format(swagger_path))
        return 2

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=swagger_path)
    if errs:
        print("Value error: Failed to parse Swagger file {}:\n{}".format(swagger_path, "\n".join(errs)))
        return 2

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)
    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    result = swagger_to.style.perform(swagger=swagger, typedefs=intermediate_typedefs, endpoints=endpoints)

    if result:
        complaints = '\n'.join(
            format_complaints(
                complaints=result,
                swagger_path=str(swagger_path),
                verbose=args.verbose,
                with_line_number=args.with_line_number))

        print("Style checks failed: \n{}".format(complaints))
        return 1

    print("Style checks succeeded.")
    return 0


def format_complaints(complaints: List[swagger_to.style.Complaint], swagger_path: str, verbose: bool,
                      with_line_number: bool) -> List[str]:
    """
    Converts a list of complaints into a well-formatted list of error messages.

    :param complaints:
    :param swagger_path:
    :param verbose:
    :param with_line_number:
    :return:
    """
    if with_line_number:
        complaints.sort(key=lambda complaint: complaint.line)

    complaints_str = []  # type: List[str]
    for complaint in complaints:
        complaint_str = ''
        if with_line_number:
            complaint_str += "{}:{} ".format(swagger_path, complaint.line)
        else:
            complaint_str += "{}: ".format(complaint.where)

        complaint_str += "{} ".format(complaint.message)

        if verbose:
            complaint_str += "\"{}\"".format(complaint.what.replace('\n', ' '))

        complaints_str.append(complaint_str)

    return complaints_str


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Generates server stubs from Swagger specification in Go.
"""

# pylint: disable=missing-docstring,too-many-instance-attributes,too-many-locals,too-many-ancestors,too-many-branches
# pylint: disable=too-many-statements, too-many-lines

import collections
import io
from typing import MutableMapping, Union, Set, List, TextIO, Optional  # pylint: disable=unused-import

import swagger_to
import swagger_to.intermediate
import swagger_to.swagger


class JsonSchema:
    def __init__(self):
        self.identifier = ''
        self.text = ''

    def text_identifier(self) -> str:
        """
        :return: name of the global variable holding the schema text
        """
        return 'jsonSchema{}Text'.format(swagger_to.capital_camel_case(self.identifier))

    def schema_identifier(self) -> str:
        """
        :return: name of the global variable holding the schema
        """
        return 'jsonSchema{}'.format(swagger_to.capital_camel_case(self.identifier))

    def validate_func_name(self) -> str:
        """
        :return: name of the function that validate the object against the schema.
        """
        return 'ValidateAgainst{}Schema'.format(swagger_to.capital_camel_case(self.identifier))


def to_json_schema(intermediate_schema: swagger_to.intermediate.JsonSchema) -> JsonSchema:
    """
    Converts the intermediate schema to a representation that we can use to easily generate go code.

    :param intermediate_schema: intermediate representation of a JSON schema
    :return: go-suitable representation
    """
    schema = JsonSchema()
    schema.identifier = intermediate_schema.identifier
    schema.text = intermediate_schema.text

    return schema


class Typedef:
    def __init__(self):
        self.identifier = ''
        self.description = ''
        self.json_schema = None  # type: Union[None, JsonSchema]


class Fielddef:
    def __init__(self):
        self.typedef = None  # type: Union[None, Typedef]
        self.description = ''
        self.json_name = ''
        self.name = ''


class Structdef(Typedef):
    def __init__(self):
        super().__init__()

        self.fields = collections.OrderedDict()  # type: MutableMapping[str, Fielddef]
        self.required = []  # type: List[str]


class Arraydef(Typedef):
    def __init__(self):
        super().__init__()

        self.items = None  # type: Union[None, Typedef]


class Mapdef(Typedef):
    def __init__(self):
        super().__init__()

        self.values = None  # type: Union[None, Typedef]


class Pointerdef(Typedef):
    def __init__(self):
        super().__init__()

        self.pointed = None  # type: Union[None, Typedef]


class Primitivedef(Typedef):
    def __init__(self):
        super().__init__()

        self.type = ''


def to_typedef(intermediate_typedef: swagger_to.intermediate.Typedef) -> Typedef:
    typedef = None  # type: Union[None, Typedef]

    if isinstance(intermediate_typedef, swagger_to.intermediate.Primitivedef):
        typedef = Primitivedef()

        if intermediate_typedef.type == 'string':
            if intermediate_typedef.format == 'date-time':
                typedef.type = 'time.Time'
            else:
                typedef.type = 'string'
        elif intermediate_typedef.type == 'number':
            if intermediate_typedef.format == '':
                typedef.type = 'float64'
            elif intermediate_typedef.format == 'float':
                typedef.type = 'float32'
            elif intermediate_typedef.format == 'double':
                typedef.type = 'float64'
            else:
                raise ValueError("Unexpected format {!r} for type {!r}".format(intermediate_typedef.format,
                                                                               intermediate_typedef.type))

        elif intermediate_typedef.type == 'integer':
            if intermediate_typedef.format == '':
                typedef.type = 'int'
            elif intermediate_typedef.format == 'int32':
                typedef.type = 'int32'
            elif intermediate_typedef.format == 'int64':
                typedef.type = 'int64'
            else:
                raise ValueError("Unexpected format {!r} for type {!r}".format(intermediate_typedef.format,
                                                                               intermediate_typedef.type))

        elif intermediate_typedef.type == 'boolean':
            typedef.type = 'bool'

        else:
            raise NotImplementedError(
                "Unhandled translation of a primitive intermediate type to Go with 'type': {!r}".format(
                    intermediate_typedef.type))

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Arraydef):
        if intermediate_typedef.items is None:
            raise ValueError("Unexpected intermediate type definition of an array to have items None: {!r}".format(
                intermediate_typedef.identifier))

        typedef = Arraydef()
        typedef.items = to_typedef(intermediate_typedef=intermediate_typedef.items)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Mapdef):
        typedef = Mapdef()
        typedef.values = to_typedef(intermediate_typedef=intermediate_typedef.values)

    elif isinstance(intermediate_typedef, swagger_to.intermediate.Objectdef):
        typedef = Structdef()

        for propdef in intermediate_typedef.properties.values():
            field = Fielddef()

            field_typedef = to_typedef(intermediate_typedef=propdef.typedef)
            if not propdef.name in intermediate_typedef.required and isinstance(field_typedef, Primitivedef):
                optional_field_typedef = Pointerdef()
                optional_field_typedef.pointed = field_typedef
                field_typedef = optional_field_typedef

            field.typedef = field_typedef
            field.description = propdef.description
            field.json_name = propdef.name
            field.name = swagger_to.capital_camel_case(identifier=propdef.name)

            typedef.fields[field.name] = field

            if propdef.name in intermediate_typedef.required:
                typedef.required.append(field.name)

    else:
        raise NotImplementedError("Unhandled translation of an intermediate type to Go: {!r}".format(
            type(intermediate_typedef)))

    assert typedef is not None

    if intermediate_typedef.identifier != '':
        typedef.identifier = swagger_to.capital_camel_case(identifier=intermediate_typedef.identifier)

    typedef.description = intermediate_typedef.description

    typedef.json_schema = to_json_schema(intermediate_schema=intermediate_typedef.json_schema)

    return typedef


def to_typedefs(
        intermediate_typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> MutableMapping[str, Typedef]:
    typedefs = collections.OrderedDict()  # type: MutableMapping[str, Typedef]

    for intermediate_typedef in intermediate_typedefs.values():
        assert intermediate_typedef is not None

        typedef = to_typedef(intermediate_typedef=intermediate_typedef)
        typedefs[typedef.identifier] = typedef

    return typedefs


def anonymous_or_get_typedef(intermediate_typedef: swagger_to.intermediate.Typedef,
                             typedefs: MutableMapping[str, Typedef]) -> Typedef:
    if intermediate_typedef.identifier != '':
        identifier = swagger_to.capital_camel_case(identifier=intermediate_typedef.identifier)

        if not identifier in typedefs:
            raise ValueError("Undefined Go type for intermediate typedef {!r}: {!r}".format(
                intermediate_typedef.identifier, identifier))

        return typedefs[identifier]

    return to_typedef(intermediate_typedef=intermediate_typedef)


def walk(typedef: Typedef, parent: Optional[Typedef]):
    yield parent, typedef

    if isinstance(typedef, Primitivedef):
        pass

    elif isinstance(typedef, Pointerdef):
        yield from walk(typedef=typedef.pointed, parent=typedef)

    elif isinstance(typedef, Arraydef):
        yield from walk(typedef=typedef.items, parent=typedef)

    elif isinstance(typedef, Mapdef):
        yield from walk(typedef=typedef.values, parent=typedef)

    elif isinstance(typedef, Structdef):
        for fielddef in typedef.fields.values():
            yield from walk(typedef=fielddef.typedef, parent=typedef)

    else:
        raise NotImplementedError("walk for Go type definition of type: {}".format(type(typedef)))


INDENT = '\t'


def write_description(description: str, fid: TextIO, indention: str) -> None:
    lines = description.strip().splitlines()
    for i, line in enumerate(lines):
        rstripped = line.rstrip()
        if len(rstripped) > 0:
            fid.write(indention + '// {}'.format(rstripped))
        else:
            fid.write(indention + '//')

        if i < len(lines) - 1:
            fid.write('\n')


def write_type_code_or_identifier(typedef: Typedef, fid: TextIO, indention: str) -> None:
    if typedef.identifier != '':
        fid.write(typedef.identifier)

    else:
        write_type_code(typedef=typedef, fid=fid, indention=indention)


def write_type_code(typedef: Typedef, fid: TextIO, indention: str) -> None:
    if isinstance(typedef, Primitivedef):
        fid.write(typedef.type)

    elif isinstance(typedef, Pointerdef):
        fid.write('*')
        write_type_code_or_identifier(typedef=typedef.pointed, fid=fid, indention=indention)

    elif isinstance(typedef, Arraydef):
        fid.write('[]')
        write_type_code_or_identifier(typedef=typedef.items, fid=fid, indention=indention)

    elif isinstance(typedef, Mapdef):
        fid.write('map[string]')
        write_type_code_or_identifier(typedef=typedef.values, fid=fid, indention=indention)

    elif isinstance(typedef, Structdef):
        fid.write('struct {\n')

        has_description = False
        for fielddef in typedef.fields.values():
            if fielddef.description != '':
                has_description = True

        for i, fielddef in enumerate(typedef.fields.values()):
            if fielddef.description != '':
                write_description(description=fielddef.description, fid=fid, indention=indention + INDENT)
                if fielddef.description != '':
                    fid.write('\n')

            fid.write(indention + INDENT)
            fid.write(fielddef.name + " ")
            write_type_code_or_identifier(typedef=fielddef.typedef, fid=fid, indention=indention + INDENT)
            fid.write(" ")

            if fielddef.name in typedef.required:
                fid.write('`json:"{}"`'.format(fielddef.json_name))
            else:
                fid.write('`json:"{},omitempty"`'.format(fielddef.json_name))

            if i < len(typedef.fields) - 1:
                fid.write("\n")

                # if at least one field has a description, add a new line for easier reading
                if has_description:
                    fid.write('\n')

        fid.write('}')

    else:
        raise NotImplementedError("No Go type writing defined for typedef of type: {!r}".format(type(typedef)))


def write_type_definition(typedef: Typedef, fid: io.StringIO) -> None:
    fid.write("type {} ".format(typedef.identifier))


def write_imports(import_set: Set[str], fid: TextIO) -> None:
    import_lst = sorted(list(import_set))

    if len(import_lst) == 1:
        fid.write('import "{}"'.format(import_lst[0]))
    elif len(import_lst) > 1:
        fid.write("import (\n")
        for import_stmt in import_lst:
            fid.write(INDENT + '"{}"\n'.format(import_stmt))
        fid.write(")")


def write_types_go(package: str, typedefs: MutableMapping[str, Typedef], fid: TextIO) -> None:
    """
    Generates a file which defines all the involved types.

    :param package: name of the package
    :param typedefs: type definitions
    :param fid: where output is written to
    :return:
    """
    fid.write("package {}\n\n".format(package))

    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    # imports
    import_set = set()  # type: Set[str]
    for typedef in typedefs.values():
        for _, another_typedef in walk(typedef=typedef, parent=None):
            if isinstance(another_typedef, Primitivedef):
                if another_typedef.type == 'time.Time':
                    import_set.add('time')

    write_imports(import_set=import_set, fid=fid)
    if len(import_set) > 0:
        fid.write('\n\n')

    # type definitions
    for i, typedef in enumerate(typedefs.values()):
        if typedef.description != '':
            write_description(description=typedef.identifier + ' ' + typedef.description, fid=fid, indention='')
            fid.write('\n')

        fid.write("type {} ".format(typedef.identifier))
        write_type_code(typedef=typedef, fid=fid, indention='')
        fid.write("\n")

        if i < len(typedefs) - 1:
            fid.write('\n')


class Argument:
    def __init__(self):
        self.typedef = None  # type: Union[None, Typedef]
        self.identifier = ''
        self.in_what = ''
        self.parameter_name = ''
        self.required = False
        self.parsing_identifier = ''
        self.json_schema = None  # type: Optional[JsonSchema]


class Handler:
    def __init__(self):
        self.identifier = ''
        self.arguments = []  # type: List[Argument]
        self.description = ''


class Wrapper:
    def __init__(self):
        self.identifier = ''
        self.description = ''
        self.handler = None  # type: Union[None, Handler]

        self.query_arguments = []  # type: List[Argument]
        self.path_arguments = []  # type: List[Argument]
        self.body_argument = None  # type: Union[None, Argument]


class Route:
    def __init__(self):
        self.description = ''
        self.path = ''
        self.method = ''
        self.wrapper = Wrapper()
        self.handler = Handler()


def endpoint_to_route_path(endpoint: swagger_to.intermediate.Endpoint) -> str:
    """
    Converts an endpoint path to Gorrila Mux route path.

    :param endpoint: whose path we need to convert
    :return: Gorrila Mux route path
    """
    token_pth = swagger_to.tokenize_path(path=endpoint.path)

    tkns = token_pth.tokens[:]

    for param in endpoint.parameters:
        if param.in_what != 'path':
            continue

        if param.name not in token_pth.parameter_to_token_indices:
            continue

        if not isinstance(param.typedef, swagger_to.intermediate.Primitivedef):
            raise ValueError("Expected path parameter {!r} in the path {!r} to be primitive, but got: {!r}".format(
                param.name, endpoint.path, type(param.typedef)))

        assert isinstance(param.typedef, swagger_to.intermediate.Primitivedef)
        if param.typedef.pattern != '':
            param_in_route = '{' + param.name + ":" + param.typedef.pattern + "}"
        else:
            param_in_route = '{' + param.name + '}'

        for i in token_pth.parameter_to_token_indices[param.name]:
            tkns[i] = param_in_route

    return "".join(tkns)


def to_route(endpoint: swagger_to.intermediate.Endpoint, typedefs: MutableMapping[str, Typedef]) -> Route:
    route = Route()
    route.method = endpoint.method.lower()
    route.path = endpoint_to_route_path(endpoint=endpoint)
    route.description = endpoint.description

    # parameters to arguments
    for param in endpoint.parameters:
        if param.in_what == 'formData':
            # no code is generated for the parameters in the form data since there are so many edge cases
            # which we possibly can't cover.
            continue
        elif param.in_what in ['query', 'body', 'path']:
            pass
        else:
            raise NotImplementedError(
                "Handling of parameters in {} is not implemented yet: end point {} {}, parameter {}.".format(
                    param.in_what, endpoint.path, endpoint.method, param.name))

        argument = Argument()
        argument.typedef = anonymous_or_get_typedef(intermediate_typedef=param.typedef, typedefs=typedefs)
        argument.required = param.required

        if not param.required and isinstance(argument.typedef, Primitivedef):
            pointer_typedef = Pointerdef()
            pointer_typedef.identifier = argument.typedef.identifier
            pointer_typedef.description = argument.typedef.description
            pointer_typedef.pointed = argument.typedef

            argument.typedef = pointer_typedef

        argument.parameter_name = param.name
        argument.parsing_identifier = swagger_to.camel_case(identifier='a_' + param.name)
        argument.identifier = swagger_to.camel_case(identifier=param.name)
        argument.in_what = param.in_what

        if param.json_schema is not None:
            argument.json_schema = to_json_schema(intermediate_schema=param.json_schema)

        if argument.in_what == 'query':
            route.wrapper.query_arguments.append(argument)

        elif argument.in_what == 'body':
            route.wrapper.body_argument = argument

        elif argument.in_what == 'path':
            route.wrapper.path_arguments.append(argument)

        else:
            raise AssertionError("Unexpected argument given in: {}".format(argument.in_what))

        route.handler.arguments.append(argument)

    route.wrapper.identifier = swagger_to.capital_camel_case(identifier='wrap_' + endpoint.operation_id)
    route.wrapper.description = '{} wraps the path `{}` with the method "{}"'.format(
        route.wrapper.identifier, route.path, route.method)
    if len(route.description) > 0:
        route.wrapper.description += '\n\nPath description:\n' + route.description

    route.wrapper.handler = route.handler
    route.handler.identifier = swagger_to.capital_camel_case(identifier=endpoint.operation_id)
    route.handler.description = '{} handles the path `{}` with the method "{}".'.format(
        route.handler.identifier, route.path, route.method)
    if len(route.description) > 0:
        route.handler.description += '\n\nPath description:\n' + route.description

    return route


def to_routes(endpoints: List[swagger_to.intermediate.Endpoint], typedefs: MutableMapping[str, Typedef]) -> List[Route]:
    routes = []  # type: List[Route]
    for endpoint in endpoints:
        routes.append(to_route(endpoint=endpoint, typedefs=typedefs))

    return routes


def write_argument_from_string(argument: Argument, string_identifier: str, indention: str, fid: TextIO) -> None:
    tajp = ''
    target_identifier = ''

    if isinstance(argument.typedef, Primitivedef):
        target_identifier = argument.parsing_identifier
        tajp = argument.typedef.type

    elif isinstance(argument.typedef, Pointerdef):
        target_identifier = argument.parsing_identifier

        if isinstance(argument.typedef.pointed, Primitivedef):
            tajp = argument.typedef.pointed.type
    else:
        raise NotImplementedError("Parsing argument from string {!r} of type: {!r}".format(
            string_identifier, type(argument)))

    if tajp == 'string':
        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + 'val := {}\n'.format(string_identifier))
            fid.write(indention + '{} = &val'.format(target_identifier))
        else:
            fid.write(indention + '{} = {}'.format(target_identifier, string_identifier))

    elif tajp == 'int':
        fid.write(indention + "{\n")
        fid.write(indention + INDENT + 'parsed, err := strconv.ParseInt({}, 10, 64)\n'.format(string_identifier))
        fid.write(indention + INDENT + "if err != nil {\n")
        fid.write(indention + INDENT * 2 + 'http.Error(w, "Parameter \'{}\': "+err.Error(), http.StatusBadRequest)\n'.
                  format(argument.parameter_name))
        fid.write(indention + INDENT * 2 + 'return\n')
        fid.write(indention + INDENT + "}\n")

        fid.write(indention + INDENT + 'converted := int(parsed)\n')

        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + INDENT + '{} = &converted\n'.format(target_identifier))
        else:
            fid.write(indention + INDENT + '{} = converted\n'.format(target_identifier))

        fid.write(indention + "}")

    elif tajp == 'int64':
        fid.write(indention + "{\n")
        fid.write(indention + INDENT + 'parsed, err := strconv.ParseInt({}, 10, 64)\n'.format(string_identifier))

        fid.write(indention + INDENT + "if err != nil {\n")
        fid.write(indention + INDENT * 2 + 'http.Error(w, "Parameter \'{}\': "+err.Error(), http.StatusBadRequest)\n'.
                  format(argument.parameter_name))
        fid.write(indention + INDENT * 2 + 'return\n')
        fid.write(indention + INDENT + "}\n")

        fid.write(indention + INDENT + 'converted := int64(parsed)\n')

        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + INDENT + '{} = &converted\n'.format(target_identifier))
        else:
            fid.write(indention + INDENT + '{} = converted\n'.format(target_identifier))

        fid.write(indention + "}")

    elif tajp == 'int32':
        fid.write(indention + "{\n")
        fid.write(indention + INDENT + 'parsed, err := strconv.ParseInt({}, 10, 32)\n'.format(string_identifier))

        fid.write(indention + INDENT + "if err != nil {\n")
        fid.write(indention + INDENT * 2 + 'http.Error(w, "Parameter \'{}\': "+err.Error(), http.StatusBadRequest)\n'.
                  format(argument.parameter_name))
        fid.write(indention + INDENT * 2 + 'return\n')
        fid.write(indention + INDENT + "}\n")

        fid.write(indention + INDENT + 'converted := int32(parsed)\n')

        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + INDENT + '{} = &converted\n'.format(target_identifier))
        else:
            fid.write(indention + INDENT + '{} = converted\n'.format(target_identifier))

        fid.write(indention + "}")

    elif tajp == 'float32':
        fid.write(indention + "{\n")
        fid.write(indention + INDENT + 'parsed, err := strconv.ParseFloat({}, 32)\n'.format(string_identifier))

        fid.write(indention + INDENT + "if err != nil {\n")
        fid.write(indention + INDENT * 2 + 'http.Error(w, "Parameter \'{}\': "+err.Error(), http.StatusBadRequest)\n'.
                  format(argument.parameter_name))
        fid.write(indention + INDENT * 2 + 'return\n')
        fid.write(indention + INDENT + "}\n")

        fid.write(indention + INDENT + 'converted := float32(parsed)\n')

        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + INDENT + '{} = &converted\n'.format(target_identifier))
        else:
            fid.write(indention + INDENT + '{} = converted\n'.format(target_identifier))

        fid.write(indention + "}")

    elif tajp == 'float64':
        fid.write(indention + "{\n")
        fid.write(indention + INDENT + 'parsed, err = strconv.ParseFloat({}, 64)\n'.format(string_identifier))

        fid.write(indention + INDENT + "if err != nil {\n")
        fid.write(indention + INDENT * 2 + 'http.Error(w, "Parameter \'{}\': "+err.Error(), http.StatusBadRequest)\n'.
                  format(argument.parameter_name))
        fid.write(indention + INDENT * 2 + 'return\n')
        fid.write(indention + INDENT + "}\n")

        fid.write(indention + INDENT + 'converted := float64(parsed)\n')

        if isinstance(argument.typedef, Pointerdef):
            fid.write(indention + INDENT + '{} = &converted\n'.format(target_identifier))
        else:
            fid.write(indention + INDENT + '{} = converted\n'.format(target_identifier))

        fid.write(indention + "}")

    else:
        raise NotImplementedError("Parsing argument from string {!r} of Go type: {!r}".format(string_identifier, tajp))


def write_parse_argument_from_body(argument: Argument, indention: str, fid: TextIO) -> None:
    fid.write(indention + "{\n")

    fid.write(indention + INDENT + "var err error\n")
    fid.write(indention + INDENT + "r.Body = http.MaxBytesReader(w, r.Body, 1024*1024)\n")
    fid.write(indention + INDENT + 'body, err := ioutil.ReadAll(r.Body)\n')
    fid.write(indention + INDENT + 'if err != nil {\n')
    fid.write(indention + 2 * INDENT + 'http.Error(w, "Body unreadable: "+err.Error(), http.StatusBadRequest)\n')
    fid.write(indention + 2 * INDENT + 'return\n')
    fid.write(indention + INDENT + '}\n\n')

    fid.write(indention + INDENT + "err = {}(body)\n".format(argument.json_schema.validate_func_name()))
    fid.write(indention + INDENT + 'if err != nil {\n')
    fid.write(indention + 2 * INDENT + 'http.Error(w, "Failed to validate against schema: "+err.Error(),\n')
    fid.write(indention + 2 * INDENT + '           http.StatusBadRequest)\n')
    fid.write(indention + 2 * INDENT + 'return\n')
    fid.write(indention + INDENT + '}\n\n')

    fid.write(indention + INDENT + "err = json.Unmarshal(body, &{})\n".format(argument.parsing_identifier))
    fid.write(indention + INDENT + 'if err != nil {\n')
    fid.write(indention + 2 * INDENT + 'http.Error(w, "Error JSON-decoding body parameter \'{}\': "+err.Error(),\n'.
              format(argument.parameter_name))
    fid.write(indention + 3 * INDENT + 'http.StatusBadRequest)\n')
    fid.write(indention + 2 * INDENT + 'return\n')
    fid.write(indention + INDENT + '}\n')

    fid.write(indention + '}')


def write_routes_go(package: str, routes: List[Route], fid: TextIO) -> None:
    """
    Generates the file which defines the router and the routes.

    :param package: name of the package
    :param routes: routes that the router will handle.
    :param fid: where the output is written to
    :return:
    """
    fid.write("package {}\n\n".format(package))
    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    # imports
    import_set = {"github.com/gorilla/mux", "net/http"}

    for route in routes:
        for argument in route.handler.arguments:
            if argument.in_what == 'body':
                import_set.add('io/ioutil')
                import_set.add('encoding/json')

            if argument.in_what == 'query':
                tajp = ''
                if isinstance(argument.typedef, Primitivedef):
                    tajp = argument.typedef.type
                elif isinstance(argument.typedef, Pointerdef):
                    if isinstance(argument.typedef.pointed, Primitivedef):
                        tajp = argument.typedef.pointed.type

                if tajp in ['int', 'int32', 'int64', 'float32', 'float64']:
                    import_set.add('strconv')

    write_imports(import_set=import_set, fid=fid)
    if len(import_set) > 0:
        fid.write('\n\n')

    # default router
    fid.write("// SetupRouter sets up a router. If you don't use any middleware, you are good to go.\n"
              "// Otherwise, you need to maually re-implement this function with your middlewares.\n"
              "func SetupRouter(h Handler) *mux.Router {\n")

    fid.write(INDENT + "r := mux.NewRouter()\n")

    for i, route in enumerate(routes):
        if i > 0:
            fid.write('\n\n')

        fid.write(INDENT + "r.HandleFunc(`{}`,\n".format(route.path))
        fid.write(INDENT * 2 + "func(w http.ResponseWriter, r *http.Request) {\n")
        fid.write(INDENT * 3 + "{}(h, w, r)\n".format(route.wrapper.identifier))
        fid.write(INDENT * 2 + '}}).Methods("{}")'.format(route.method))

    if len(routes) > 0:
        fid.write('\n')
    fid.write(INDENT + 'return r\n')

    fid.write("}\n")

    if len(routes) > 0:
        fid.write("\n")

    # wrappers
    for route_i, route in enumerate(routes):
        if route_i > 0:
            fid.write('\n')

        wrapper = route.wrapper
        if len(wrapper.description):
            write_description(description=wrapper.description, fid=fid, indention='')
            fid.write('\n')

        fid.write('func {}(h Handler, w http.ResponseWriter, r *http.Request) {{\n'.format(wrapper.identifier))

        # intermidiate variables
        for argument in route.handler.arguments:
            fid.write(INDENT + 'var {} '.format(argument.parsing_identifier))
            write_type_code_or_identifier(typedef=argument.typedef, fid=fid, indention=INDENT)
            fid.write('\n')

        # query arguments
        if len(route.wrapper.query_arguments) > 0:
            fid.write('\n')
            fid.write(INDENT + "q := r.URL.Query()\n\n")

        for i, argument in enumerate(route.wrapper.query_arguments):
            if i > 0:
                fid.write("\n")

            if argument.required:
                fid.write(INDENT + 'if _, ok := q["{}"]; !ok {{\n'.format(argument.parameter_name))
                fid.write(INDENT * 2 + 'http.Error(w, "Parameter \'{}\' expected in query", http.StatusBadRequest)\n'.
                          format(argument.parameter_name))
                fid.write(INDENT * 2 + 'return\n')
                fid.write(INDENT + "}\n")
                write_argument_from_string(
                    argument=argument,
                    string_identifier='q.Get("{}")'.format(argument.parameter_name),
                    indention=INDENT,
                    fid=fid)
                fid.write('\n')
            else:
                fid.write(INDENT + 'if _, ok := q["{}"]; ok {{\n'.format(argument.parameter_name))
                write_argument_from_string(
                    argument=argument,
                    string_identifier='q.Get("{}")'.format(argument.parameter_name),
                    indention=INDENT * 2,
                    fid=fid)
                fid.write('\n')
                fid.write(INDENT + '}\n')

        # path arguments
        if len(route.wrapper.path_arguments) > 0:
            fid.write("\n")
            fid.write(INDENT + "vars := mux.Vars(r)\n")

        for i, argument in enumerate(route.wrapper.path_arguments):
            if i > 0:
                fid.write("\n")

            fid.write("\n")
            if argument.required:
                fid.write(INDENT + 'if _, ok := vars["{}"]; !ok {{\n'.format(argument.parameter_name))
                fid.write(INDENT * 2 + 'http.Error(w, "Parameter \'{}\' expected in path", http.StatusBadRequest)\n'.
                          format(argument.parameter_name))
                fid.write(INDENT * 2 + 'return\n')
                fid.write(INDENT + "}\n")
                write_argument_from_string(
                    argument=argument,
                    string_identifier='vars["{}"]'.format(argument.parameter_name),
                    indention=INDENT,
                    fid=fid)
                fid.write('\n')
            else:
                fid.write(INDENT + 'if value, ok := vars["{}"]; ok {{\n'.format(argument.parameter_name))
                write_argument_from_string(argument=argument, string_identifier='value', indention=INDENT * 2, fid=fid)
                fid.write('\n}\n')

        # body argument
        if route.wrapper.body_argument is not None:
            if len(route.wrapper.path_arguments) > 0 or len(route.wrapper.query_arguments) > 0:
                fid.write('\n')

            if route.wrapper.body_argument.required:
                fid.write(INDENT + 'if r.Body == nil {\n')
                fid.write(INDENT * 2 + (
                    'http.Error(w, "Parameter \'{}\' expected in body, '
                    'but got no body", http.StatusBadRequest)\n').format(route.wrapper.body_argument.parameter_name))

                fid.write(INDENT * 2 + 'return\n')
                fid.write(INDENT + '}\n')

                write_parse_argument_from_body(argument=route.wrapper.body_argument, indention=INDENT, fid=fid)
                fid.write('\n')
            else:
                fid.write(INDENT + 'if r.Body != nil {\n')
                write_parse_argument_from_body(argument=route.wrapper.body_argument, indention=2 * INDENT, fid=fid)
                fid.write('\n')
                fid.write(INDENT + '}\n')

        # call handler
        if len(route.wrapper.query_arguments) > 0 or \
                len(route.wrapper.path_arguments) > 0 or route.wrapper.body_argument is not None:
            fid.write("\n")
        prefix = INDENT + 'h.{}('.format(route.handler.identifier)
        fid.write(prefix)

        fid.write('w,\n')
        fid.write(2 * INDENT + 'r')

        for argument in route.handler.arguments:
            fid.write(",\n")
            fid.write(2 * INDENT + argument.parsing_identifier)
        fid.write(")")

        fid.write('\n}\n')

    fid.write("\n// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")


def write_handler_impl_go(package: str, routes: List[Route], fid: TextIO) -> None:
    """
    Generates the file which implements the handler interface with empty methods.

    :param package: name of the package
    :param routes: that a handler will handle
    :param fid: where the output is written to
    :return:
    """
    fid.write("package {}\n\n".format(package))

    fid.write('import (\n')
    fid.write(INDENT + '"net/http"\n')
    fid.write(INDENT + '"log"\n')
    fid.write(')\n\n')

    # yapf: disable
    fid.write('// HandlerImpl implements the Handler.\n')
    fid.write('type HandlerImpl struct {\n' +
              INDENT + 'LogErr *log.Logger\n' +
              INDENT + 'LogOut *log.Logger}')
    # yapf: enable

    if routes:
        fid.write('\n\n')

    for i, route in enumerate(routes):
        if i > 0:
            fid.write('\n\n')

        fid.write('// {0} implements Handler.{0}.\n'.format(route.handler.identifier))
        fid.write("func (h *HandlerImpl) {}(".format(route.handler.identifier))
        fid.write('w http.ResponseWriter,\n')
        fid.write(INDENT + 'r *http.Request')

        for argument in route.handler.arguments:
            fid.write(",\n")
            fid.write(INDENT)

            fid.write(argument.identifier + ' ')
            write_type_code_or_identifier(typedef=argument.typedef, fid=fid, indention='')

        fid.write(") {\n")
        fid.write(INDENT + 'http.Error(w, "Not implemented: {}", http.StatusInternalServerError)\n'.format(
            route.handler.identifier))
        fid.write(INDENT + 'h.LogErr.Printf("Not implemented: {}")\n'.format(route.handler.identifier))
        fid.write("}")

    fid.write('\n')


def write_handler_go(package: str, routes: List[Route], fid: TextIO) -> None:
    """
    Generates the file which defines the handler interface.

    :param package: name of the package
    :param routes: that a handler will handle
    :param fid: where the output is written to
    :return:
    """
    fid.write("package {}\n\n".format(package))

    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    fid.write('import "net/http"\n\n')

    fid.write('// Handler defines an interface to handling the routes.\n')
    fid.write('type Handler interface {\n')

    for i, route in enumerate(routes):
        if i > 0:
            fid.write('\n')

        if len(route.handler.description) > 0:
            write_description(description=route.handler.description, fid=fid, indention=INDENT)
            fid.write('\n')

        fid.write(INDENT + "{}(".format(route.handler.identifier))
        fid.write('w http.ResponseWriter,\n')
        fid.write(2 * INDENT + 'r *http.Request')

        for argument in route.handler.arguments:
            fid.write(",\n")
            fid.write(2 * INDENT)

            fid.write(argument.identifier + ' ')
            write_type_code_or_identifier(typedef=argument.typedef, fid=fid, indention='')

        fid.write(")\n")

    fid.write('}\n')

    fid.write("\n// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")


def write_json_schemas_go(package: str, routes: List[Route], typedefs: MutableMapping[str, Typedef],
                          fid: TextIO) -> None:
    """
    Represents the definitions as json schemas and hard-codes them as strings in Go.

    It is assumed that the Swagger definitions already represent a subset of JSON Schema.
    This is theoretically not the case (some formats are swagger-only), but in most cases
    the literal translation should work.

    :param package: package name
    :param routes: needed to generate the parameter schemas if they are not already defined in the definitions
    :param typedefs: type definitions to generate the schemas for
    :param fid: where output is written to
    """
    fid.write("package {}\n\n".format(package))

    fid.write("// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n\n")

    schemas = collections.OrderedDict()  # type: MutableMapping[str, JsonSchema]

    for route in routes:
        if route.wrapper.body_argument is not None:
            if not route.wrapper.body_argument.json_schema.identifier in schemas:
                schemas[route.wrapper.body_argument.json_schema.identifier] = route.wrapper.body_argument.json_schema

    for typedef_id, typedef in typedefs.items():
        if typedef.json_schema is None:
            raise AssertionError("Missing JSON schema for typedef: {!r}".format(typedef_id))

        # Assume the typedef identifiers are unique accross routes and typedefs.
        if typedef.json_schema.identifier not in schemas:
            schemas[typedef.json_schema.identifier] = typedef.json_schema

    if len(schemas) == 0:
        return

    fid.write('import (\n')
    fid.write(INDENT + '"errors"\n')
    fid.write(INDENT + '"fmt"\n')
    fid.write(INDENT + '"github.com/xeipuuv/gojsonschema"\n')
    fid.write(")\n\n")

    fid.write("func mustNewJSONSchema(text string, name string) *gojsonschema.Schema {\n")
    fid.write(INDENT + "loader := gojsonschema.NewStringLoader(text)\n")
    fid.write(INDENT + "schema, err := gojsonschema.NewSchema(loader)\n")
    fid.write(INDENT + "if err != nil {\n")
    fid.write(INDENT * 2 + 'panic(fmt.Sprintf("failed to load JSON Schema %#v: %s", text, err.Error()))\n')
    fid.write(INDENT + "}\n")
    fid.write(INDENT + "return schema\n")
    fid.write("}\n\n")

    for i, schema in enumerate(schemas.values()):
        if i > 0:
            fid.write('\n\n')

        fid.write('var {} = `{}`'.format(schema.text_identifier(), schema.text.replace(r'`', '` + "`" + `')))

    fid.write('\n\n')

    for schema in schemas.values():
        fid.write('var {} = mustNewJSONSchema({}, "{}")\n'.format(schema.schema_identifier(), schema.text_identifier(),
                                                                  schema.identifier))

    fid.write('\n')

    for i, schema in enumerate(schemas.values()):
        if i > 0:
            fid.write('\n\n')

        fid.write("// {} validates a message coming from the client against {} schema.\n".format(
            schema.validate_func_name(), schema.identifier))

        fid.write("func {}(bb []byte) error {{\n".format(schema.validate_func_name()))
        fid.write(INDENT + "loader := gojsonschema.NewStringLoader(string(bb))\n")
        fid.write(INDENT + "result, err := {}.Validate(loader)\n".format(schema.schema_identifier()))
        fid.write(INDENT + "if err != nil {\n")
        fid.write(2 * INDENT + "return err\n")
        fid.write(INDENT + "}\n\n")
        fid.write(INDENT + "if result.Valid() {\n")
        fid.write(2 * INDENT + "return nil\n")
        fid.write(INDENT + "}\n\n")

        fid.write(INDENT + 'msg := ""\n')
        fid.write(INDENT + 'for i, valErr := range result.Errors() {\n')
        fid.write(2 * INDENT + 'if i > 0 {\n')
        fid.write(3 * INDENT + 'msg += ", "\n')
        fid.write(2 * INDENT + '}\n')
        fid.write(2 * INDENT + 'msg += valErr.String()\n')
        fid.write(INDENT + '}\n')
        fid.write(INDENT + "return errors.New(msg)\n")
        fid.write("}")

    fid.write("\n\n// Automatically generated file by swagger_to. DO NOT EDIT OR APPEND ANYTHING!\n")



#!/usr/bin/env python3
"""
Reads a correct swagger file and produces Typescript client code.
"""
import argparse
import os

import swagger_to.intermediate
import swagger_to.swagger
import swagger_to.ts_angular5_client


def main() -> None:
    """"
    Main routine
    """
    parser = argparse.ArgumentParser("Reads a correct swagger file and produces Typescript client code")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--outpath", help="path to the output file", required=True)
    parser.add_argument("--force", help="overwrite existing file", action="store_true")
    args = parser.parse_args()

    swagger_path = str(args.swagger_path)
    out_path = str(args.outpath)
    force = bool(args.force)

    if not os.path.exists(args.swagger_path):
        raise FileNotFoundError("Swagger file does not exist: {}".format(swagger_path))

    os.makedirs(os.path.dirname(out_path), exist_ok=True)

    if not force and os.path.exists(out_path):
        raise FileExistsError("Output path already exists and --force was not specified: {}".format(out_path))

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=args.swagger_path)
    if errs:
        raise AssertionError("Failed to parse Swagger file {!r}:\n{}".format(swagger_path, "\n".join(errs)))

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)
    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    ts_typedefs = swagger_to.ts_angular5_client.to_typedefs(intermediate_typedefs=intermediate_typedefs)
    ts_requests = swagger_to.ts_angular5_client.to_requests(endpoints=endpoints, typedefs=ts_typedefs)

    with open(out_path, 'wt') as fid:
        swagger_to.ts_angular5_client.write_client_ts(typedefs=ts_typedefs, requests=ts_requests, fid=fid)

    print("Generated Typescript + Angular5 client code in: {}".format(out_path))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Reads a correct swagger file and produces Go code
"""
import argparse
import os
import sys

import swagger_to.go_server
import swagger_to.intermediate
import swagger_to.swagger


def main() -> None:
    """"
    Main routine
    """
    parser = argparse.ArgumentParser("Reads a correct swagger file and produces Go code")
    parser.add_argument("--swagger_path", help="path to the swagger file", required=True)
    parser.add_argument("--outdir", help="path to the output directory", required=True)
    parser.add_argument("--force", help="overwrite existing files", action="store_true")
    args = parser.parse_args()

    assert isinstance(args.force, bool)
    assert isinstance(args.outdir, str)
    assert isinstance(args.swagger_path, str)

    if not os.path.exists(args.swagger_path):
        print("Swagger file does not exist: {}".format(args.swagger_path), file=sys.stderr)
        sys.exit(1)

    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    if not args.force:
        for fname in ['main.go', 'handlers.go', 'types.go', 'routes.go', 'jsonschemas.go']:
            pth = os.path.join(args.outdir, fname)
            if os.path.exists(pth):
                print("File exists, but --force was not specified: {!r}".format(pth))
                sys.exit(1)

    swagger, errs = swagger_to.swagger.parse_yaml_file(path=args.swagger_path)
    if errs:
        raise AssertionError("Errors in {}:\n{}".format(args.swagger_path, "\n".join(errs)))

    intermediate_typedefs = swagger_to.intermediate.to_typedefs(swagger=swagger)
    intermediate_params = swagger_to.intermediate.to_parameters(swagger=swagger, typedefs=intermediate_typedefs)

    endpoints = swagger_to.intermediate.to_endpoints(
        swagger=swagger, typedefs=intermediate_typedefs, params=intermediate_params)

    go_typedefs = swagger_to.go_server.to_typedefs(intermediate_typedefs=intermediate_typedefs)
    go_routes = swagger_to.go_server.to_routes(endpoints=endpoints, typedefs=go_typedefs)

    package = swagger.name

    pth = os.path.join(args.outdir, 'types.go')
    with open(pth, 'wt') as fid:
        swagger_to.go_server.write_types_go(package=package, typedefs=go_typedefs, fid=fid)

    pth = os.path.join(args.outdir, 'routes.go')
    with open(pth, 'wt') as fid:
        swagger_to.go_server.write_routes_go(package=package, routes=go_routes, fid=fid)

    pth = os.path.join(args.outdir, 'handler.go')
    with open(pth, 'wt') as fid:
        swagger_to.go_server.write_handler_go(package=package, routes=go_routes, fid=fid)

    pth = os.path.join(args.outdir, 'handler_impl.go.sample')
    with open(pth, 'wt') as fid:
        swagger_to.go_server.write_handler_impl_go(package=package, routes=go_routes, fid=fid)

    pth = os.path.join(args.outdir, 'jsonschemas.go')
    with open(pth, 'wt') as fid:
        swagger_to.go_server.write_json_schemas_go(package=package, routes=go_routes, typedefs=go_typedefs, fid=fid)

    print("Generated go server code in: {}".format(args.outdir))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Generates code for an Elm client.
"""
# pylint: disable=too-many-lines
from typing import MutableSet, MutableMapping, List, Optional

import swagger_to.intermediate
import swagger_to.swagger
import swagger_to


class Complaint:
    """Encapsulates a complaint."""

    def __init__(self, message: str, what: str, where: str, line: int) -> None:
        self.message = message
        self.what = what
        self.where = where
        self.line = line


def check_header(swagger: swagger_to.swagger.Swagger) -> List[Complaint]:
    """
    Checks whether the swagger header conforms to the style guide.

    :param swagger: parsed swagger file
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    if swagger.name != swagger_to.snake_case(swagger.name):
        complaints.append(
            Complaint(
                message="Name of the Swagger specification is not snake case (e.g. snake_case)",
                what=swagger.name,
                where="In the Swagger header",
                line=1))

    if not swagger.base_path.startswith("/"):
        complaints.append(
            Complaint(
                message="Swagger base path doesn't start with a slash",
                what=swagger.base_path,
                where="In the Swagger header",
                line=1))

    if swagger.description.capitalize() != swagger.description:
        complaints.append(
            Complaint(
                message="Swagger description should be capitalized",
                what=swagger.description,
                where="In the Swagger header",
                line=1))

    return complaints


def check_casing_typedefs(typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> List[Complaint]:
    """
    Checks whether the typedefs conform to the casing conventions.

    :param typedefs: swagger type definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    seen_typedefs = set()  # type: MutableSet
    for _, typedef in enumerate(typedefs.values()):
        complaints.extend(_check_recursively_cases(typedef=typedef, visited=seen_typedefs))

    return complaints


def check_casing_endpoints(endpoints: List[swagger_to.intermediate.Endpoint]) -> List[Complaint]:
    """
    Checks whether the endpoints conform to the casing conventions.

    :param endpoints: swagger endpoint definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    for endpoint in endpoints:
        if endpoint.path != swagger_to.snake_case(endpoint.path):
            complaints.append(
                Complaint(
                    message="Path doesn't conform to snake case (e.g. snake_case)",
                    what=endpoint.path,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))
        if endpoint.operation_id != swagger_to.snake_case(endpoint.operation_id):
            complaints.append(
                Complaint(
                    message="Endpoint operation ID is not a snake case identifier (e.g. snake_case)",
                    what=endpoint.operation_id,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))
        for param in endpoint.parameters:
            if param.name != swagger_to.snake_case(param.name):
                complaints.append(
                    Complaint(
                        message="Parameter has not a snake case identifier (e.g. snake_case)",
                        what=param.name,
                        where="In endpoint {}, parameter {}".format(endpoint.operation_id, param.name),
                        line=param.line))

    return complaints


def check_descriptions_typedefs(typedefs: MutableMapping[str, swagger_to.intermediate.Typedef]) -> List[Complaint]:
    """
    Checks whether the typedefs conform to the description conventions.

    :param typedefs: swagger type definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    seen_typedefs = set()  # type: MutableSet
    for _, typedef in enumerate(typedefs.values()):
        complaints.extend(_check_recursively_descriptions(typedef=typedef, visited=seen_typedefs))

    return complaints


def check_descriptions_endpoints(endpoints: List[swagger_to.intermediate.Endpoint]) -> List[Complaint]:
    """
    Checks whether the endpoints conform to the description conventions.

    :param endpoints: swagger endpoint definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    for endpoint in endpoints:
        if check_description(endpoint.description):
            complaints.append(
                Complaint(
                    message=check_description(endpoint.description),
                    what=endpoint.description,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))
        for param in endpoint.parameters:
            if check_description(param.description):
                complaints.append(
                    Complaint(
                        message=check_description(param.description),
                        what=param.description,
                        where="In endpoint {}, parameter {}".format(endpoint.operation_id, param.name),
                        line=param.line))
        for _, resp in enumerate(endpoint.responses.values()):
            if check_description(resp.description):
                complaints.append(
                    Complaint(
                        message=check_description(resp.description),
                        what=resp.description,
                        where="In endpoint {}, response {}".format(endpoint.operation_id, resp.code),
                        line=resp.line))

    return complaints


def _check_recursively_cases(typedef: swagger_to.intermediate.Typedef,
                             visited: MutableSet[swagger_to.intermediate.Typedef]) -> List[Complaint]:
    """
    Checks the typedef's adherence to the casing conventions.

    :param typedef: to be translated
    :param visited: already seen typedefs
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]
    if typedef in visited:
        return complaints

    visited.add(typedef)

    if isinstance(typedef, swagger_to.intermediate.Primitivedef):
        pass

    elif isinstance(typedef, swagger_to.intermediate.Arraydef):
        if typedef.identifier != "" and typedef.identifier != swagger_to.capital_camel_case(typedef.identifier):
            complaints.append(
                Complaint(
                    message="Not a capital camel case identifier (e.g. CamelCase)",
                    what=typedef.identifier,
                    where="In array {}".format(typedef.identifier),
                    line=typedef.line))
        complaints.extend(_check_recursively_cases(typedef=typedef.items, visited=visited))

    elif isinstance(typedef, swagger_to.intermediate.Mapdef):
        if typedef.identifier != "" and typedef.identifier != swagger_to.capital_camel_case(typedef.identifier):
            complaints.append(
                Complaint(
                    message="Not a capital camel case identifier (e.g. CamelCase)",
                    what=typedef.identifier,
                    where="In map {}".format(typedef.identifier),
                    line=typedef.line))
        complaints.extend(_check_recursively_cases(typedef=typedef.values, visited=visited))

    elif isinstance(typedef, swagger_to.intermediate.Objectdef):

        if typedef.identifier != "" and typedef.identifier != swagger_to.capital_camel_case(typedef.identifier):
            complaints.append(
                Complaint(
                    message="Not a capital camel case identifier (e.g. CamelCase)",
                    what=typedef.identifier,
                    where="In object {}".format(typedef.identifier),
                    line=typedef.line))

        for prop in typedef.properties.values():
            if prop.name != swagger_to.snake_case(prop.name):
                complaints.append(
                    Complaint(
                        message="Not a snake case identifier (e.g. snake_case)",
                        what=prop.name,
                        where="In object {}, property {}".format(typedef.identifier, prop.name),
                        line=typedef.line))
            complaints.extend(_check_recursively_cases(typedef=prop.typedef, visited=visited))

    return complaints


def _check_recursively_descriptions(typedef: swagger_to.intermediate.Typedef,
                                    visited: MutableSet[swagger_to.intermediate.Typedef]) -> List[Complaint]:
    """
    Checks the typedef's adherence to the description conventions.

    :param typedef: to be translated
    :param visited: already seen typedefs
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]
    if typedef in visited:
        return complaints

    visited.add(typedef)

    if isinstance(typedef, swagger_to.intermediate.Primitivedef):
        pass

    elif isinstance(typedef, swagger_to.intermediate.Arraydef):
        if check_description(typedef.description):
            complaints.append(
                Complaint(
                    message=check_description(typedef.description),
                    what=typedef.description,
                    where="In array {}".format(typedef.identifier),
                    line=typedef.line))
        complaints.extend(_check_recursively_descriptions(typedef=typedef.items, visited=visited))

    elif isinstance(typedef, swagger_to.intermediate.Mapdef):
        if check_description(typedef.description):
            complaints.append(
                Complaint(
                    what=typedef.description,
                    message=check_description(typedef.description),
                    where="In map {}".format(typedef.identifier),
                    line=typedef.line))
        complaints.extend(_check_recursively_descriptions(typedef=typedef.values, visited=visited))

    elif isinstance(typedef, swagger_to.intermediate.Objectdef):
        if check_description(typedef.description):
            complaints.append(
                Complaint(
                    what=typedef.description,
                    message=check_description(typedef.description),
                    where="In object {}".format(typedef.identifier),
                    line=typedef.line))

        for prop in typedef.properties.values():
            if check_description(prop.description):
                complaints.append(
                    Complaint(
                        what=prop.description,
                        message=check_description(prop.description),
                        where="In object {}, property {}".format(typedef.identifier, prop.name),
                        line=typedef.line))
                complaints.extend(_check_recursively_descriptions(typedef=prop.typedef, visited=visited))

    return complaints


def check_endpoint_responses(endpoints: List[swagger_to.intermediate.Endpoint]) -> List[Complaint]:
    """
    Checks whether the endpoints conform to the conventions for responses.

    :param endpoints: swagger endpoint definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    for endpoint in endpoints:
        if "200" not in endpoint.responses.keys():
            complaints.append(
                Complaint(
                    message="Path doesn't include response 200",
                    what=endpoint.path,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))
        if "default" not in endpoint.responses.keys():
            complaints.append(
                Complaint(
                    message="Path doesn't include default response",
                    what=endpoint.path,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))

    return complaints


def check_endpoint_path(endpoints: List[swagger_to.intermediate.Endpoint]) -> List[Complaint]:
    """
    Checks whether the endpoints conform to the conventions for paths.

    :param endpoints: swagger endpoint definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    for endpoint in endpoints:
        if not endpoint.path.startswith("/"):
            complaints.append(
                Complaint(
                    message="Path doesn't begin with a slash",
                    what=endpoint.path,
                    where="In endpoint {}".format(endpoint.operation_id),
                    line=endpoint.line))

    return complaints


def check_description(description: str) -> Optional[str]:
    """
    Checks whether a description is well-styled.
    :param description: the description
    :return: the failed check, if any
    """
    # pylint: disable=too-many-return-statements, too-many-branches
    if description == "":
        return None

    without_pipe = description.replace("|", "")

    if not without_pipe[:1].isalpha():
        return "description should start with alphanumeric character"

    if without_pipe[:1].isupper():
        return "description should start with lower case character"

    words = without_pipe.split(' ')

    if not words[0].endswith('s'):
        return "description should start with verb in present tense (stem + \"-s\")"

    lines = without_pipe.splitlines()

    if not lines[0].endswith('.'):
        return "description's first line should end with a period"

    one_empty_line = False
    for i, line in enumerate(lines):
        if line.strip() != line:
            return "no line in the description should contain leading or trailing white space"

        if i == 1 and line.strip() != "":
            return "description's second line should be empty"

        if line == "":
            if one_empty_line:
                return "description should not contain two empty lines"
            else:
                one_empty_line = True
        else:
            one_empty_line = False

    if without_pipe.strip() != without_pipe:
        return "description should not contain leading or trailing whitespaces"

    return None


def perform(swagger: swagger_to.swagger.Swagger, typedefs: MutableMapping[str, swagger_to.intermediate.Typedef],
            endpoints: List[swagger_to.intermediate.Endpoint]) -> List[Complaint]:
    """
    Checks whether the typedefs and endpoints conform to the swagger style guide.

    :param swagger: parsed swagger file
    :param typedefs: swagger type definitions
    :param endpoints: swagger endpoint definitions
    :return: the list of failed checks
    """
    complaints = []  # type: List[Complaint]

    complaints.extend(check_header(swagger=swagger))

    complaints.extend(check_casing_typedefs(typedefs=typedefs))

    complaints.extend(check_casing_endpoints(endpoints=endpoints))

    complaints.extend(check_descriptions_typedefs(typedefs=typedefs))

    complaints.extend(check_descriptions_endpoints(endpoints=endpoints))

    complaints.extend(check_endpoint_path(endpoints=endpoints))

    complaints.extend(check_endpoint_responses(endpoints=endpoints))

    complaints.sort(key=lambda complaint: complaint.where)
    return complaints
